# mypy: disable-error-code="method-assign,no-untyped-def"

import pytest
from rdflib import RDF
from rdflib import RDFS
from rdflib import SH
from rdflib import XSD
from rdflib import Graph
from rdflib import Literal

from dsp_tools.commands.validate_data.models.input_problems import ProblemType
from dsp_tools.commands.validate_data.models.input_problems import Severity
from dsp_tools.commands.validate_data.models.validation import DetailBaseInfo
from dsp_tools.commands.validate_data.models.validation import QueryInfo
from dsp_tools.commands.validate_data.models.validation import UnexpectedComponent
from dsp_tools.commands.validate_data.models.validation import ValidationReportGraphs
from dsp_tools.commands.validate_data.models.validation import ValidationResult
from dsp_tools.commands.validate_data.models.validation import ValidationResultBaseInfo
from dsp_tools.commands.validate_data.models.validation import ViolationType
from dsp_tools.commands.validate_data.process_validation_report.query_validation_result import (
    _extract_base_info_of_resource_results,
)
from dsp_tools.commands.validate_data.process_validation_report.query_validation_result import _get_all_main_result_bns
from dsp_tools.commands.validate_data.process_validation_report.query_validation_result import (
    _get_resource_iri_and_type,
)
from dsp_tools.commands.validate_data.process_validation_report.query_validation_result import _query_all_results
from dsp_tools.commands.validate_data.process_validation_report.query_validation_result import _query_one_with_detail
from dsp_tools.commands.validate_data.process_validation_report.query_validation_result import _query_one_without_detail
from dsp_tools.commands.validate_data.process_validation_report.query_validation_result import (
    _reformat_one_validation_result,
)
from dsp_tools.commands.validate_data.process_validation_report.query_validation_result import _separate_result_types
from dsp_tools.commands.validate_data.process_validation_report.query_validation_result import reformat_validation_graph
from dsp_tools.utils.rdflib_constants import DATA
from dsp_tools.utils.rdflib_constants import KNORA_API
from test.unittests.commands.validate_data.constants import IN_BUILT_ONTO
from test.unittests.commands.validate_data.constants import ONTO


def test_reformat_validation_graph(report_target_resource_wrong_type: tuple[Graph, Graph]) -> None:
    validation_g, onto_data_g = report_target_resource_wrong_type
    report = ValidationReportGraphs(
        conforms=False,
        validation_graph=validation_g,
        shacl_graph=Graph(),
        onto_graph=onto_data_g,
        data_graph=onto_data_g,
    )
    result_all_problems = reformat_validation_graph(report)
    assert not result_all_problems.unexpected_results
    assert len(result_all_problems.problems) == 1
    result = result_all_problems.problems.pop(0)
    assert result.problem_type == ProblemType.LINK_TARGET_TYPE_MISMATCH
    assert result.res_id == "region_isRegionOf_resource_not_a_representation"
    assert result.res_type == "Region"
    assert result.prop_name == "isRegionOf"
    assert result.severity == Severity.VIOLATION
    assert result.input_value == "target_res_without_representation_1"
    assert result.input_type == "in-built:TestNormalResource"
    assert result.expected == "Representation"


def test_separate_bns_of_results(
    report_target_resource_wrong_type: tuple[Graph, Graph], report_not_resource: tuple[Graph, Graph]
):
    val_g1, _ = report_target_resource_wrong_type
    val_g2, _ = report_not_resource
    combined_g = val_g1 + val_g2
    extracted_bns = _get_all_main_result_bns(combined_g)
    node1 = next(combined_g.subjects(SH.focusNode, DATA.region_isRegionOf_resource_not_a_representation))
    node2 = next(combined_g.subjects(SH.focusNode, DATA.value_id_simpletext))
    assert extracted_bns == {node1, node2}


class TestGetResourceIRIs:
    def test_with_detail_user_facing_info_there(self, report_value_type_simpletext):
        _, onto_data_g, base_info = report_value_type_simpletext
        query_info = QueryInfo(base_info.result_bn, DATA.id_simpletext, ONTO.ClassWithEverything)
        resource_iri, resource_type, user_facing_prop = _get_resource_iri_and_type(
            query_info, ONTO.testTextarea, onto_data_g, {KNORA_API.TextValue}
        )
        assert resource_iri == DATA.id_simpletext
        assert resource_type == ONTO.ClassWithEverything
        assert user_facing_prop == ONTO.testTextarea

    def test_no_detail_user_facing_info_there(self, report_value_type):
        _, onto_data_g, base_info = report_value_type
        query_info = QueryInfo(base_info.result_bn, DATA.id_uri, ONTO.ClassWithEverything)
        resource_iri, resource_type, user_facing_prop = _get_resource_iri_and_type(
            query_info, ONTO.testUriValue, onto_data_g, {KNORA_API.TextValue}
        )
        assert resource_iri == DATA.id_uri
        assert resource_type == ONTO.ClassWithEverything
        assert user_facing_prop == ONTO.testUriValue

    def test_no_detail_user_facing_prop_is_knora_prop(self, report_archive_missing_legal_info):
        validation_g, onto_data_g = report_archive_missing_legal_info
        result_bn = next(validation_g.subjects(RDF.type, SH.ValidationResult))
        query_info = QueryInfo(result_bn, DATA.value_bitstream_no_legal_info, KNORA_API.ArchiveFileValue)
        resource_iri, resource_type, user_facing_prop = _get_resource_iri_and_type(
            query_info, KNORA_API.hasLicense, onto_data_g, {KNORA_API.ArchiveFileValue}
        )
        assert resource_iri == DATA.bitstream_no_legal_info
        assert resource_type == ONTO.TestArchiveRepresentation
        assert user_facing_prop == KNORA_API.hasLicense


class TestQueryAllResults:
    def test_link_target_inexistent(self, report_target_resource_wrong_type: tuple[Graph, Graph]) -> None:
        validation_g, onto_data_g = report_target_resource_wrong_type
        extracted_results, unexpected_components = _query_all_results(validation_g, onto_data_g)
        assert not unexpected_components
        assert len(extracted_results) == 1
        result = extracted_results.pop(0)
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.LINK_TARGET
        assert result.res_iri == DATA.region_isRegionOf_resource_not_a_representation
        assert result.res_class == KNORA_API.Region
        assert result.property == KNORA_API.isRegionOf
        assert result.severity == SH.Violation
        assert result.input_value == DATA.target_res_without_representation_1
        assert result.input_type == IN_BUILT_ONTO.TestNormalResource
        assert result.expected == Literal("http://api.knora.org/ontology/knora-api/v2#Representation")

    def test_report_archive_missing_legal_info(self, report_archive_missing_legal_info: tuple[Graph, Graph]) -> None:
        validation_g, onto_data_g = report_archive_missing_legal_info
        extracted_results, unexpected_components = _query_all_results(validation_g, onto_data_g)
        assert not unexpected_components
        assert len(extracted_results) == 1
        result = extracted_results.pop(0)
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.GENERIC
        assert result.res_iri == DATA.bitstream_no_legal_info
        assert result.res_class == ONTO.TestArchiveRepresentation
        assert result.property == KNORA_API.hasLicense
        assert result.severity == SH.Warning
        assert not result.input_value
        assert not result.input_type
        assert result.expected == Literal("Files and IIIF-URIs require a reference to a license.")

    def test_result_geoname_not_number(self, report_regex: tuple[Graph, Graph, ValidationResultBaseInfo]) -> None:
        res, data, _ = report_regex
        extracted_results, unexpected_components = _query_all_results(res, data)
        assert not unexpected_components
        assert len(extracted_results) == 1
        result = extracted_results.pop(0)
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.PATTERN
        assert result.severity == SH.Violation
        assert result.res_iri == DATA.geoname_not_number
        assert result.res_class == ONTO.ClassWithEverything
        assert result.property == ONTO.testGeoname
        assert result.expected == Literal("The value must be a valid geoname code")
        assert result.input_value == Literal("this-is-not-a-valid-code")


class TestExtractBaseInfo:
    def test_no_detail(self, report_min_card: tuple[Graph, Graph, ValidationResultBaseInfo]) -> None:
        validation_g, onto_data_g, _ = report_min_card
        results = _extract_base_info_of_resource_results(validation_g, onto_data_g)
        assert len(results) == 1
        found_result = results[0]
        assert found_result.focus_node_iri == DATA.id_card_one
        assert found_result.focus_node_type == ONTO.ClassInheritedCardinalityOverwriting
        assert found_result.result_path == ONTO.testBoolean
        assert found_result.source_constraint_component == SH.MinCountConstraintComponent
        assert found_result.severity == SH.Violation
        assert not found_result.detail

    def test_still_image_file(
        self, report_image_missing_legal_info: tuple[Graph, Graph, ValidationResultBaseInfo]
    ) -> None:
        validation_g, onto_data_g, _ = report_image_missing_legal_info
        results = _extract_base_info_of_resource_results(validation_g, onto_data_g)
        assert len(results) == 1
        found_result = results[0]
        assert found_result.focus_node_iri == DATA.image_no_legal_info
        assert found_result.focus_node_type == ONTO.TestStillImageRepresentation
        assert found_result.result_path == KNORA_API.hasLicense
        assert found_result.source_constraint_component == SH.MinCountConstraintComponent
        assert found_result.severity == SH.Warning
        assert not found_result.detail

    def test_with_detail(self, report_value_type_simpletext: tuple[Graph, Graph, ValidationResultBaseInfo]) -> None:
        validation_g, onto_data_g, _ = report_value_type_simpletext
        results = _extract_base_info_of_resource_results(validation_g, onto_data_g)
        assert len(results) == 1
        found_result = results[0]
        assert found_result.focus_node_iri == DATA.id_simpletext
        assert found_result.focus_node_type == ONTO.ClassWithEverything
        assert found_result.result_path == ONTO.testTextarea
        assert found_result.source_constraint_component == SH.NodeConstraintComponent
        assert found_result.severity == SH.Violation
        detail = found_result.detail
        assert isinstance(detail, DetailBaseInfo)
        assert detail.source_constraint_component == SH.MinCountConstraintComponent


class TestSeparateResultTypes:
    def test_result_id_card_one(self, report_min_card: tuple[Graph, Graph, ValidationResultBaseInfo]) -> None:
        res_g, onto_data_g, _ = report_min_card
        no_detail, with_detail = _separate_result_types(res_g, onto_data_g)
        assert len(no_detail) == 1
        assert len(with_detail) == 0
        assert no_detail[0].focus_node_iri == DATA.id_card_one

    def test_result_id_simpletext(
        self, report_value_type_simpletext: tuple[Graph, Graph, ValidationResultBaseInfo]
    ) -> None:
        res_g, onto_data_g, _ = report_value_type_simpletext
        no_detail, with_detail = _separate_result_types(res_g, onto_data_g)
        assert len(no_detail) == 0
        assert len(with_detail) == 1
        assert with_detail[0].focus_node_iri == DATA.id_simpletext

    def test_result_id_uri(self, report_value_type: tuple[Graph, Graph, ValidationResultBaseInfo]) -> None:
        res_g, onto_data_g, _ = report_value_type
        no_detail, with_detail = _separate_result_types(res_g, onto_data_g)
        assert len(no_detail) == 1
        assert len(with_detail) == 0
        assert no_detail[0].focus_node_iri == DATA.id_uri

    def test_result_geoname_not_number(self, report_regex: tuple[Graph, Graph, ValidationResultBaseInfo]) -> None:
        res_g, onto_data_g, _ = report_regex
        no_detail, with_detail = _separate_result_types(res_g, onto_data_g)
        assert len(no_detail) == 1
        assert len(with_detail) == 0
        assert no_detail[0].focus_node_iri == DATA.geoname_not_number

    def test_result_id_closed_constraint(
        self, report_closed_constraint: tuple[Graph, Graph, ValidationResultBaseInfo]
    ) -> None:
        res_g, onto_data_g, _ = report_closed_constraint
        no_detail, with_detail = _separate_result_types(res_g, onto_data_g)
        assert len(no_detail) == 1
        assert len(with_detail) == 0
        assert no_detail[0].focus_node_iri == DATA.id_closed_constraint

    def test_result_id_max_card(self, report_max_card: tuple[Graph, Graph, ValidationResultBaseInfo]) -> None:
        res_g, onto_data_g, _ = report_max_card
        no_detail, with_detail = _separate_result_types(res_g, onto_data_g)
        assert len(no_detail) == 1
        assert len(with_detail) == 0
        assert no_detail[0].focus_node_iri == DATA.id_max_card

    def test_report_unique_value_literal(
        self, report_unique_value_literal: tuple[Graph, Graph, ValidationResultBaseInfo]
    ) -> None:
        res_g, onto_data_g, _ = report_unique_value_literal
        no_detail, with_detail = _separate_result_types(res_g, onto_data_g)
        assert len(no_detail) == 1
        assert len(with_detail) == 0
        assert no_detail[0].focus_node_iri == DATA.identical_values_valueHas

    def test_report_unique_value_iri(
        self, report_unique_value_iri: tuple[Graph, Graph, ValidationResultBaseInfo]
    ) -> None:
        res_g, onto_data_g, _ = report_unique_value_iri
        no_detail, with_detail = _separate_result_types(res_g, onto_data_g)
        assert len(no_detail) == 1
        assert len(with_detail) == 0
        assert no_detail[0].focus_node_iri == DATA.identical_values_LinkValue


class TestQueryWithoutDetail:
    def test_result_id_card_one(self, report_min_card: tuple[Graph, Graph, ValidationResultBaseInfo]) -> None:
        res, data, info = report_min_card
        result = _query_one_without_detail(info, res, data)
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.MIN_CARD
        assert result.res_iri == info.focus_node_iri
        assert result.res_class == info.focus_node_type
        assert result.property == ONTO.testBoolean
        assert result.severity == SH.Violation
        assert result.expected == Literal("1")

    def test_result_id_closed_constraint(
        self, report_closed_constraint: tuple[Graph, Graph, ValidationResultBaseInfo]
    ) -> None:
        res, data, info = report_closed_constraint
        result = _query_one_without_detail(info, res, data)
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.NON_EXISTING_CARD
        assert result.res_iri == info.focus_node_iri
        assert result.res_class == info.focus_node_type
        assert result.severity == SH.Violation
        assert result.property == ONTO.testIntegerSimpleText

    def test_result_id_max_card(self, report_max_card: tuple[Graph, Graph, ValidationResultBaseInfo]) -> None:
        res, data, info = report_max_card
        result = _query_one_without_detail(info, res, data)
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.MAX_CARD
        assert result.res_iri == info.focus_node_iri
        assert result.res_class == info.focus_node_type
        assert result.property == ONTO.testHasLinkToCardOneResource
        assert result.severity == SH.Violation
        assert result.expected == Literal("1")

    def test_result_empty_label(self, report_empty_label: tuple[Graph, ValidationResultBaseInfo]) -> None:
        graphs, info = report_empty_label
        result = _query_one_without_detail(info, graphs, Graph())
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.PATTERN
        assert result.res_iri == info.focus_node_iri
        assert result.res_class == info.focus_node_type
        assert result.property == RDFS.label
        assert result.severity == SH.Violation
        assert result.expected == Literal("The label must be a non-empty string")
        assert result.input_value == Literal(" ")

    def test_unique_value_literal(
        self, report_unique_value_literal: tuple[Graph, Graph, ValidationResultBaseInfo]
    ) -> None:
        res, data, info = report_unique_value_literal
        result = _query_one_without_detail(info, res, data)
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.UNIQUE_VALUE
        assert result.res_iri == info.focus_node_iri
        assert result.res_class == info.focus_node_type
        assert result.severity == SH.Violation
        assert result.property == ONTO.testGeoname
        assert result.input_value == Literal("00111111")

    def test_unique_value_iri(self, report_unique_value_iri: tuple[Graph, Graph, ValidationResultBaseInfo]) -> None:
        res, data, info = report_unique_value_iri
        result = _query_one_without_detail(info, res, data)
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.UNIQUE_VALUE
        assert result.res_iri == info.focus_node_iri
        assert result.res_class == info.focus_node_type
        assert result.severity == SH.Violation
        assert result.property == ONTO.testHasLinkTo
        assert result.input_value == DATA.link_valueTarget_id

    def test_coexist_with(self, report_coexist_with: tuple[Graph, Graph, ValidationResultBaseInfo]) -> None:
        validation_g, data, info = report_coexist_with
        result = _query_one_without_detail(info, validation_g, data)
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.SEQNUM_IS_PART_OF
        assert result.res_iri == info.focus_node_iri
        assert result.res_class == info.focus_node_type
        assert result.severity == SH.Violation
        assert result.message == Literal("The property seqnum must be used together with isPartOf")
        assert not result.property
        assert not result.input_value

    def test_report_coexist_with_date(
        self, report_coexist_with_date: tuple[Graph, Graph, ValidationResultBaseInfo]
    ) -> None:
        validation_g, data, info = report_coexist_with_date
        result = _query_one_without_detail(info, validation_g, data)
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.GENERIC
        assert result.res_iri == info.focus_node_iri
        assert result.res_class == info.focus_node_type
        assert result.severity == SH.Violation
        assert result.message == Literal("date message")
        assert result.property == info.result_path
        assert result.input_value == Literal("GREGORIAN:CE:2000:BCE:1900", datatype=XSD.string)

    def test_image_missing_legal_info(
        self, report_image_missing_legal_info: tuple[Graph, Graph, ValidationResultBaseInfo]
    ) -> None:
        res, data, info = report_image_missing_legal_info
        result = _query_one_without_detail(info, res, data)
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.GENERIC
        assert result.res_iri == info.focus_node_iri
        assert result.res_class == info.focus_node_type
        assert result.severity == SH.Warning
        assert result.property == KNORA_API.hasLicense
        assert result.expected == Literal("Files and IIIF-URIs require a reference to a license.")

    def test_result_id_uri(self, report_value_type: tuple[Graph, Graph, ValidationResultBaseInfo]) -> None:
        res, data, info = report_value_type
        result = _query_one_without_detail(info, res, data)
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.VALUE_TYPE
        assert result.res_iri == info.focus_node_iri
        assert result.res_class == info.focus_node_type
        assert result.property == ONTO.testUriValue
        assert result.severity == SH.Violation
        assert result.expected == Literal("This property requires a UriValue")
        assert result.input_type == KNORA_API.TextValue

    def test_report_min_inclusive(self, report_min_inclusive: tuple[Graph, Graph, ValidationResultBaseInfo]) -> None:
        res, data, info = report_min_inclusive
        result = _query_one_without_detail(info, res, data)
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.GENERIC
        assert result.res_iri == info.focus_node_iri
        assert result.res_class == info.focus_node_type
        assert result.property == KNORA_API.hasSegmentBounds
        assert result.severity == SH.Violation
        assert result.message == Literal("The interval start must be a non-negative integer or decimal.")
        assert result.input_value == Literal("-2.0", datatype=XSD.decimal)

    def test_report_single_line_constraint_component(self, report_single_line_constraint_component) -> None:
        res, data, info = report_single_line_constraint_component
        result = _query_one_without_detail(info, res, data)
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.GENERIC
        assert result.res_iri == info.focus_node_iri
        assert result.res_class == info.focus_node_type
        assert result.severity == SH.Violation
        assert result.property == KNORA_API.hasCopyrightHolder
        assert result.message == Literal("The copyright holder must be a string without newlines.")
        assert result.input_value == Literal(
            """FirstLine
Second Line"""
        )

    def test_report_date_single_month_does_not_exist(
        self, report_date_single_month_does_not_exist: tuple[Graph, Graph, ValidationResultBaseInfo]
    ) -> None:
        res, data, info = report_date_single_month_does_not_exist
        result = _query_one_without_detail(info, res, data)
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.GENERIC
        assert result.res_iri == info.focus_node_iri
        assert result.res_class == info.focus_node_type
        assert result.property == ONTO.testSubDate1
        assert result.severity == SH.Violation
        assert result.message == Literal("date message")
        assert result.input_value == Literal("GREGORIAN:CE:1800-22", datatype=XSD.string)

    def test_report_date_range_wrong_yyyy(
        self, report_date_range_wrong_yyyy: tuple[Graph, Graph, ValidationResultBaseInfo]
    ) -> None:
        res, data, info = report_date_range_wrong_yyyy
        result = _query_one_without_detail(info, res, data)
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.GENERIC
        assert result.res_iri == info.focus_node_iri
        assert result.res_class == info.focus_node_type
        assert result.property == ONTO.testSubDate1
        assert result.severity == SH.Violation
        assert result.message == Literal("date message")
        assert result.input_value == Literal("GREGORIAN:CE:2000:CE:1900", datatype=XSD.string)

    def test_report_date_range_wrong_to_ignore(
        self, report_date_range_wrong_to_ignore: tuple[Graph, Graph, ValidationResultBaseInfo]
    ) -> None:
        res, data, info = report_date_range_wrong_to_ignore
        result = _query_one_without_detail(info, res, data)
        assert not result

    def test_unknown(self, result_unknown_component: tuple[Graph, ValidationResultBaseInfo]) -> None:
        graphs, info = result_unknown_component
        result = _query_one_without_detail(info, graphs, Graph())
        assert isinstance(result, UnexpectedComponent)
        assert result.component_type == str(SH.UniqueLangConstraintComponent)


class TestQueryWithDetail:
    def test_result_id_simpletext(
        self, report_value_type_simpletext: tuple[Graph, Graph, ValidationResultBaseInfo]
    ) -> None:
        res, data, info = report_value_type_simpletext
        result = _query_one_with_detail(info, res, data)
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.VALUE_TYPE
        assert result.res_iri == info.focus_node_iri
        assert result.res_class == info.focus_node_type
        assert result.property == ONTO.testTextarea
        assert result.severity == SH.Violation
        assert result.expected == Literal("TextValue without formatting")
        assert result.input_type == KNORA_API.TextValue

    def test_link_target_non_existent(
        self, report_link_target_non_existent: tuple[Graph, Graph, ValidationResultBaseInfo]
    ) -> None:
        res, data, info = report_link_target_non_existent
        result = _query_one_with_detail(info, res, data)
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.LINK_TARGET
        assert result.res_iri == info.focus_node_iri
        assert result.res_class == info.focus_node_type
        assert result.property == ONTO.testHasLinkTo
        assert result.severity == SH.Violation
        assert result.expected == KNORA_API.Resource
        assert result.input_value == DATA.other
        assert not result.input_type

    def test_link_target_wrong_class(
        self, report_link_target_wrong_class: tuple[Graph, Graph, ValidationResultBaseInfo]
    ) -> None:
        res, data, info = report_link_target_wrong_class
        result = _query_one_with_detail(info, res, data)
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.LINK_TARGET
        assert result.res_iri == info.focus_node_iri
        assert result.res_class == info.focus_node_type
        assert result.severity == SH.Violation
        assert result.property == ONTO.testHasLinkToCardOneResource
        assert result.expected == ONTO.CardOneResource
        assert result.input_value == DATA.id_9_target
        assert result.input_type == ONTO.ClassWithEverything

    def test_report_unknown_list_name(
        self, report_unknown_list_name: tuple[Graph, Graph, ValidationResultBaseInfo]
    ) -> None:
        res, data, info = report_unknown_list_name
        result = _query_one_with_detail(info, res, data)
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.GENERIC
        assert result.res_iri == info.focus_node_iri
        assert result.res_class == info.focus_node_type
        assert result.severity == SH.Violation
        assert result.property == ONTO.testListProp
        assert result.message == Literal("A valid node from the list 'firstList' must be used with this property.")
        assert result.input_value == Literal("other / n1")

    def test_report_unknown_list_node(
        self, report_unknown_list_node: tuple[Graph, Graph, ValidationResultBaseInfo]
    ) -> None:
        res, data, info = report_unknown_list_node
        result = _query_one_with_detail(info, res, data)
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.GENERIC
        assert result.res_iri == info.focus_node_iri
        assert result.res_class == info.focus_node_type
        assert result.severity == SH.Violation
        assert result.property == ONTO.testListProp
        assert result.message == Literal("A valid node from the list 'firstList' must be used with this property.")
        assert result.input_value == Literal("firstList / other")

    def test_report_single_line_constraint_component_content_is_value(
        self, report_single_line_constraint_component_content_is_value
    ) -> None:
        res, data, info = report_single_line_constraint_component_content_is_value
        result = _query_one_with_detail(info, res, data)
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.GENERIC
        assert result.res_iri == info.focus_node_iri
        assert result.res_class == info.focus_node_type
        assert result.severity == SH.Violation
        assert result.property == ONTO.testSimpleText
        assert result.message == Literal("The value must be a non-empty string without newlines.")
        assert result.input_value == Literal(
            """This may not

have newlines"""
        )


class TestQueryFileValueViolations:
    def test_missing_file_value(self, report_missing_file_value: tuple[Graph, ValidationResultBaseInfo]) -> None:
        graphs, info = report_missing_file_value
        result = _query_one_without_detail(info, graphs, Graph())
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.MIN_CARD
        assert result.res_iri == info.focus_node_iri
        assert result.res_class == info.focus_node_type
        assert result.severity == SH.Violation
        assert result.property == KNORA_API.hasMovingImageFileValue
        assert result.expected == Literal("Cardinality 1")

    def test_report_file_closed_constraint(
        self, report_file_closed_constraint: tuple[Graph, Graph, ValidationResultBaseInfo]
    ) -> None:
        results_g, data, info = report_file_closed_constraint
        result = _query_one_without_detail(info, results_g, data)
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.FILE_VALUE_PROHIBITED
        assert result.res_iri == info.focus_node_iri
        assert result.res_class == info.focus_node_type
        assert result.severity == SH.Violation
        assert result.property == KNORA_API.hasMovingImageFileValue
        assert result.input_value == Literal("file.mp4", datatype=XSD.string)

    def test_file_value_for_resource_without_representation(
        self, file_value_for_resource_without_representation: tuple[Graph, ValidationResultBaseInfo]
    ) -> None:
        graphs, info = file_value_for_resource_without_representation
        result = _query_one_without_detail(info, graphs, Graph())
        assert isinstance(result, ValidationResult)
        assert result.violation_type == ViolationType.FILE_VALUE_PROHIBITED
        assert result.res_iri == info.focus_node_iri
        assert result.res_class == info.focus_node_type
        assert result.severity == SH.Violation
        assert result.property == KNORA_API.hasMovingImageFileValue


class TestReformatResult:
    def test_min(self, extracted_min_card: ValidationResult) -> None:
        result = _reformat_one_validation_result(extracted_min_card)
        assert result.problem_type == ProblemType.MIN_CARD
        assert result.res_id == "id_card_one"
        assert result.res_type == "onto:ClassInheritedCardinalityOverwriting"
        assert result.prop_name == "onto:testBoolean"
        assert result.severity == Severity.VIOLATION
        assert result.expected == "1"

    def test_max(self, extracted_max_card: ValidationResult) -> None:
        result = _reformat_one_validation_result(extracted_max_card)
        assert result.problem_type == ProblemType.MAX_CARD
        assert result.res_id == "id_max_card"
        assert result.res_type == "onto:ClassMixedCard"
        assert result.prop_name == "onto:testDecimalSimpleText"
        assert result.severity == Severity.VIOLATION
        assert result.expected == "0-1"

    def test_violation_empty_label(self, extracted_empty_label: ValidationResult) -> None:
        result = _reformat_one_validation_result(extracted_empty_label)
        assert result.problem_type == ProblemType.INPUT_REGEX
        assert result.res_id == "empty_label"
        assert result.res_type == "onto:ClassWithEverything"
        assert result.prop_name == "rdfs:label"
        assert result.severity == Severity.VIOLATION
        assert result.expected == "The label must be a non-empty string"
        assert result.input_value == " "

    def test_closed(self, extracted_closed_constraint: ValidationResult) -> None:
        result = _reformat_one_validation_result(extracted_closed_constraint)
        assert result.problem_type == ProblemType.NON_EXISTING_CARD
        assert result.res_id == "id_closed_constraint"
        assert result.res_type == "onto:CardOneResource"
        assert result.prop_name == "onto:testIntegerSimpleText"
        assert result.severity == Severity.VIOLATION

    def test_value_type_simpletext(self, extracted_value_type_simpletext: ValidationResult) -> None:
        result = _reformat_one_validation_result(extracted_value_type_simpletext)
        assert result.problem_type == ProblemType.VALUE_TYPE_MISMATCH
        assert result.res_id == "id_simpletext"
        assert result.res_type == "onto:ClassWithEverything"
        assert result.prop_name == "onto:testTextarea"
        assert result.severity == Severity.VIOLATION
        assert result.input_type == "TextValue"
        assert result.expected == "TextValue without formatting"

    def test_value_type(self, extracted_value_type: ValidationResult) -> None:
        result = _reformat_one_validation_result(extracted_value_type)
        assert result.problem_type == ProblemType.VALUE_TYPE_MISMATCH
        assert result.res_id == "id_uri"
        assert result.res_type == "onto:ClassWithEverything"
        assert result.prop_name == "onto:testUriValue"
        assert result.severity == Severity.VIOLATION
        assert result.input_type == "TextValue"
        assert result.expected == "This property requires a UriValue"

    def test_violation_regex(self, extracted_regex: ValidationResult) -> None:
        result = _reformat_one_validation_result(extracted_regex)
        assert result.problem_type == ProblemType.INPUT_REGEX
        assert result.res_id == "geoname_not_number"
        assert result.res_type == "onto:ClassWithEverything"
        assert result.prop_name == "onto:testGeoname"
        assert result.severity == Severity.VIOLATION
        assert result.expected == "The value must be a valid geoname code"
        assert result.input_value == "this-is-not-a-valid-code"

    def test_link_target_non_existent(self, extracted_link_target_non_existent: ValidationResult) -> None:
        result = _reformat_one_validation_result(extracted_link_target_non_existent)
        assert result.problem_type == ProblemType.INEXISTENT_LINKED_RESOURCE
        assert result.res_id == "link_target_non_existent"
        assert result.res_type == "onto:ClassWithEverything"
        assert result.prop_name == "onto:testHasLinkTo"
        assert result.severity == Severity.VIOLATION
        assert result.input_value == "other"

    def test_link_target_wrong_class(self, extracted_link_target_wrong_class: ValidationResult) -> None:
        result = _reformat_one_validation_result(extracted_link_target_wrong_class)
        assert result.problem_type == ProblemType.LINK_TARGET_TYPE_MISMATCH
        assert result.res_id == "link_target_wrong_class"
        assert result.res_type == "onto:ClassWithEverything"
        assert result.prop_name == "onto:testHasLinkToCardOneResource"
        assert result.severity == Severity.VIOLATION
        assert result.input_value == "id_9_target"
        assert result.input_type == "onto:ClassWithEverything"
        assert result.expected == "onto:CardOneResource"

    def test_unique_value_literal(self, extracted_unique_value_literal: ValidationResult) -> None:
        result = _reformat_one_validation_result(extracted_unique_value_literal)
        assert result.problem_type == ProblemType.DUPLICATE_VALUE
        assert result.res_id == "identical_values_valueHas"
        assert result.res_type == "onto:ClassWithEverything"
        assert result.prop_name == "onto:testGeoname"
        assert result.severity == Severity.VIOLATION
        assert result.input_value == "00111111"

    def test_unique_value_iri(self, extracted_unique_value_iri: ValidationResult) -> None:
        result = _reformat_one_validation_result(extracted_unique_value_iri)
        assert result.problem_type == ProblemType.DUPLICATE_VALUE
        assert result.res_id == "identical_values_LinkValue"
        assert result.res_type == "onto:ClassWithEverything"
        assert result.prop_name == "onto:testHasLinkTo"
        assert result.severity == Severity.VIOLATION
        assert result.input_value == "link_valueTarget_id"

    def test_unknown_list_node(self, extracted_unknown_list_node: ValidationResult) -> None:
        result = _reformat_one_validation_result(extracted_unknown_list_node)
        assert result.problem_type == ProblemType.GENERIC
        assert result.res_id == "list_node_non_existent"
        assert result.res_type == "onto:ClassWithEverything"
        assert result.prop_name == "onto:testListProp"
        assert result.severity == Severity.VIOLATION
        assert result.message == "A valid node from the list 'firstList' must be used with this property."
        assert result.input_value == "firstList / other"

    def test_unknown_list_name(self, extracted_unknown_list_name: ValidationResult) -> None:
        result = _reformat_one_validation_result(extracted_unknown_list_name)
        assert result.problem_type == ProblemType.GENERIC
        assert result.res_id == "list_name_non_existent"
        assert result.res_type == "onto:ClassWithEverything"
        assert result.prop_name == "onto:testListProp"
        assert result.severity == Severity.VIOLATION
        assert result.message == "A valid node from the list 'firstList' must be used with this property."
        assert result.input_value == "other / n1"

    def test_min_inclusive(self, extracted_min_inclusive: ValidationResult) -> None:
        result = _reformat_one_validation_result(extracted_min_inclusive)
        assert result.problem_type == ProblemType.GENERIC
        assert result.res_id == "video_segment_negative_bounds"
        assert result.res_type == "VideoSegment"
        assert result.prop_name == "hasSegmentBounds"
        assert result.severity == Severity.VIOLATION
        assert result.message == "The interval start must be a non-negative integer or decimal."
        assert result.input_value == "-2.0"

    def test_missing_file_value(self, extracted_missing_file_value: ValidationResult) -> None:
        result = _reformat_one_validation_result(extracted_missing_file_value)
        assert result.problem_type == ProblemType.FILE_VALUE_MISSING
        assert result.res_id == "id_video_missing"
        assert result.res_type == "onto:TestMovingImageRepresentation"
        assert result.prop_name == "bitstream"
        assert result.severity == Severity.VIOLATION
        assert result.expected == "This resource requires a file with one of the following extensions: 'mp4'"

    def test_image_missing_legal_info(self, extracted_image_missing_legal_info: ValidationResult) -> None:
        result = _reformat_one_validation_result(extracted_image_missing_legal_info)
        assert result.problem_type == ProblemType.GENERIC
        assert result.res_id == "image_no_legal_info"
        assert result.res_type == "onto:TestStillImageRepresentation"
        assert result.prop_name == "bitstream / iiif-uri"
        assert result.severity == Severity.WARNING
        assert result.expected == "Files and IIIF-URIs require a reference to a license."

    def test_single_line_constraint_component(
        self, extracted_single_line_constraint_component: ValidationResult
    ) -> None:
        result = _reformat_one_validation_result(extracted_single_line_constraint_component)
        assert result.problem_type == ProblemType.GENERIC
        assert result.res_id == "copyright_holder_with_newline"
        assert result.res_type == "onto:TestArchiveRepresentation"
        assert result.prop_name == "bitstream / iiif-uri"
        assert result.severity == Severity.VIOLATION
        assert result.message == "The copyright holder must be a string without newlines."
        assert result.input_value == "with newline"

    def test_file_value_for_resource_without_representation(
        self, extracted_file_value_for_resource_without_representation: ValidationResult
    ) -> None:
        result = _reformat_one_validation_result(extracted_file_value_for_resource_without_representation)
        assert result.problem_type == ProblemType.FILE_VALUE_PROHIBITED
        assert result.res_id == "id_resource_without_representation"
        assert result.res_type == "onto:ClassWithEverything"
        assert result.prop_name == "bitstream / iiif-uri"
        assert result.severity == Severity.VIOLATION

    def test_seqnum_is_part_of(self, extracted_coexist_with: ValidationResult) -> None:
        result = _reformat_one_validation_result(extracted_coexist_with)
        assert result.problem_type == ProblemType.GENERIC
        assert result.res_id == "missing_seqnum"
        assert result.res_type == "in-built:TestStillImageRepresentationWithSeqnum"
        assert result.prop_name == "seqnum or isPartOf"
        assert result.severity == Severity.VIOLATION
        assert result.message == "Coexist message from knora-api turtle"

    def test_date_single_month_does_not_exist(
        self, extracted_date_single_month_does_not_exist: ValidationResult
    ) -> None:
        result = _reformat_one_validation_result(extracted_date_single_month_does_not_exist)
        assert result.problem_type == ProblemType.GENERIC
        assert result.res_id == "date_month_does_not_exist"
        assert result.res_type == "onto:ClassWithEverything"
        assert result.prop_name == "onto:testSubDate1"
        assert result.severity == Severity.VIOLATION
        assert result.message == "date message"
        assert result.input_value == "1800-22"


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

from enum import Enum
from enum import unique
from typing import Optional


@unique
class PermissionValue(Enum):
    RV = 1
    V = 2
    M = 4
    D = 8
    CR = 16

    def __str__(self) -> str:
        tmp = {
            1: "RV",
            2: "V",
            4: "M",
            8: "D",
            16: "CR",
        }
        return tmp[self.value]


class Permissions:
    _permissions: dict[PermissionValue, list[str]]

    def __init__(self, permissions: Optional[dict[PermissionValue, list[str]]] = None):
        self._permissions = permissions or {}

    def __str__(self) -> str:
        tmpstr = ""
        for permission, groups in self._permissions.items():
            if tmpstr:
                tmpstr += "|"
            tmpstr += f"{permission!s} " + ",".join(groups)
        return tmpstr

    def add(self, key: PermissionValue, val: str) -> None:
        if self._permissions.get(key) is None:
            self._permissions[key] = [val]
        else:
            self._permissions[key].append(val)



# mypy: disable-error-code="method-assign,no-untyped-def"

import pytest

from dsp_tools.commands.validate_data.models.api_responses import ListLookup
from dsp_tools.commands.validate_data.models.rdf_like_data import PropertyObject
from dsp_tools.commands.validate_data.models.rdf_like_data import RdfLikeResource
from dsp_tools.commands.validate_data.models.rdf_like_data import RdfLikeValue
from dsp_tools.commands.validate_data.models.rdf_like_data import TripleObjectType
from dsp_tools.commands.validate_data.models.rdf_like_data import TriplePropertyType
from dsp_tools.commands.validate_data.prepare_data.get_rdf_like_data import _get_date_str
from dsp_tools.commands.validate_data.prepare_data.get_rdf_like_data import _get_file_metadata
from dsp_tools.commands.validate_data.prepare_data.get_rdf_like_data import _get_file_value
from dsp_tools.commands.validate_data.prepare_data.get_rdf_like_data import _get_list_value_str
from dsp_tools.commands.validate_data.prepare_data.get_rdf_like_data import _get_one_resource
from dsp_tools.commands.validate_data.prepare_data.get_rdf_like_data import _get_one_value
from dsp_tools.commands.validate_data.prepare_data.get_rdf_like_data import _get_xsd_like_dates
from dsp_tools.commands.validate_data.prepare_data.get_rdf_like_data import get_rdf_like_data
from dsp_tools.utils.data_formats.date_util import Era
from dsp_tools.utils.data_formats.date_util import SingleDate
from dsp_tools.utils.rdflib_constants import KNORA_API_STR
from dsp_tools.utils.xml_parsing.models.parsed_resource import KnoraValueType
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedFileValue
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedFileValueMetadata
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedResource
from dsp_tools.utils.xml_parsing.models.parsed_resource import ParsedValue

ONTO = "http://0.0.0.0:3333/ontology/9999/onto/v2#"
HAS_PROP = f"{ONTO}hasProp"
RES_TYPE = f"{ONTO}ResourceType"


AUTHORSHIP_LOOKUP = {"auth_id": ["Author Resolved"]}
LIST_LOOKUP = ListLookup({("list", "node"): "http://rdfh.ch/lists/9999/n1"})


@pytest.fixture
def file_with_permission() -> ParsedFileValue:
    metadata = ParsedFileValueMetadata("http://rdfh.ch/licenses/cc-by-nc-4.0", "copy", "auth_id", "public")
    return ParsedFileValue("file.jpg", KnoraValueType.STILL_IMAGE_FILE, metadata)


@pytest.fixture
def bool_value() -> ParsedValue:
    return ParsedValue(HAS_PROP, "true", KnoraValueType.BOOLEAN_VALUE, None, None)


@pytest.fixture
def richtext_with_standoff() -> ParsedValue:
    text_str = 'With <a class="salsah-link" href="IRI:link:IRI">link text</a>.'
    return ParsedValue(HAS_PROP, text_str, KnoraValueType.RICHTEXT_VALUE, None, None)


def _get_label_and_type(resource: RdfLikeResource) -> tuple[PropertyObject, PropertyObject, list[PropertyObject]]:
    lbl = next(x for x in resource.property_objects if x.property_type == TriplePropertyType.RDFS_LABEL)
    rdf_type = next(x for x in resource.property_objects if x.property_type == TriplePropertyType.RDF_TYPE)
    permissions = list(x for x in resource.property_objects if x.property_type == TriplePropertyType.KNORA_PERMISSIONS)
    return lbl, rdf_type, permissions


class TestResource:
    def test_empty(self):
        res = ParsedResource(
            res_id="one",
            res_type=RES_TYPE,
            label="lbl",
            permissions_id=None,
            values=[],
            file_value=None,
            migration_metadata=None,
        )
        result = _get_one_resource(res, {}, LIST_LOOKUP)
        assert result.res_id == "one"
        assert len(result.property_objects) == 2
        lbl, rdf_type, _ = _get_label_and_type(result)
        assert lbl.object_value == "lbl"
        assert lbl.object_type == TripleObjectType.STRING
        assert rdf_type.object_value == RES_TYPE
        assert rdf_type.object_type == TripleObjectType.IRI
        assert len(result.values) == 0

    def test_empty_permissions(self):
        res = ParsedResource(
            res_id="one",
            res_type=RES_TYPE,
            label="lbl",
            permissions_id="public",
            values=[],
            file_value=None,
            migration_metadata=None,
        )
        result = _get_one_resource(res, {}, LIST_LOOKUP)
        assert result.res_id == "one"
        assert len(result.property_objects) == 3
        lbl, rdf_type, perm = _get_label_and_type(result)
        assert len(perm) == 1
        permission = perm.pop(0)
        assert permission.object_value == "public"
        assert permission.object_type == TripleObjectType.STRING
        assert permission.property_type == TriplePropertyType.KNORA_PERMISSIONS
        assert lbl.object_value == "lbl"
        assert lbl.object_type == TripleObjectType.STRING
        assert rdf_type.object_value == RES_TYPE
        assert rdf_type.object_type == TripleObjectType.IRI
        assert len(result.values) == 0

    def test_with_props(self, bool_value):
        res = ParsedResource(
            res_id="one",
            res_type=RES_TYPE,
            label="lbl",
            permissions_id=None,
            values=[bool_value],
            file_value=None,
            migration_metadata=None,
        )
        result_list = get_rdf_like_data([res], {}, LIST_LOOKUP)
        assert len(result_list.resources) == 1
        result = result_list.resources.pop(0)
        assert result.res_id == "one"
        assert len(result.property_objects) == 2
        lbl, rdf_type, _ = _get_label_and_type(result)
        assert lbl.object_value == "lbl"
        assert lbl.object_type == TripleObjectType.STRING
        assert rdf_type.object_value == RES_TYPE
        assert rdf_type.object_type == TripleObjectType.IRI
        assert len(result.values) == 1

    def test_with_file_value(self, file_with_permission):
        res = ParsedResource(
            res_id="one",
            res_type=RES_TYPE,
            label="lbl",
            permissions_id=None,
            values=[],
            file_value=file_with_permission,
            migration_metadata=None,
        )
        result = _get_one_resource(res, AUTHORSHIP_LOOKUP, LIST_LOOKUP)
        assert len(result.values) == 1
        file_value = result.values.pop(0)
        assert isinstance(file_value, RdfLikeValue)
        assert file_value.user_facing_prop == f"{KNORA_API_STR}hasStillImageFileValue"
        assert file_value.user_facing_value == "file.jpg"
        assert file_value.knora_type == KnoraValueType.STILL_IMAGE_FILE
        assert len(file_value.value_metadata) == 4

    def test_with_standoff(self, richtext_with_standoff):
        res = ParsedResource(
            res_id="one",
            res_type=RES_TYPE,
            label="lbl",
            permissions_id=None,
            values=[richtext_with_standoff],
            file_value=None,
            migration_metadata=None,
        )
        result = _get_one_resource(res, {}, LIST_LOOKUP)
        assert result.res_id == "one"
        assert len(result.property_objects) == 3
        lbl, rdf_type, _ = _get_label_and_type(result)
        assert lbl.object_value == "lbl"
        assert lbl.object_type == TripleObjectType.STRING
        assert rdf_type.object_value == RES_TYPE
        assert rdf_type.object_type == TripleObjectType.IRI
        standoff = next(x for x in result.property_objects if x.property_type == TriplePropertyType.KNORA_STANDOFF_LINK)
        assert standoff.object_value == "link"
        assert standoff.object_type == TripleObjectType.IRI
        assert len(result.values) == 1


class TestValues:
    def test_boolean_corr(self):
        val = ParsedValue(HAS_PROP, "true", KnoraValueType.BOOLEAN_VALUE, None, None)
        res = _get_one_value(val, LIST_LOOKUP)
        assert res.user_facing_prop == HAS_PROP
        assert res.user_facing_value == "true"
        assert res.knora_type == KnoraValueType.BOOLEAN_VALUE
        assert not res.value_metadata

    def test_boolean_with_comment_corr(self):
        val = ParsedValue(HAS_PROP, "true", KnoraValueType.BOOLEAN_VALUE, None, "comment")
        res = _get_one_value(val, LIST_LOOKUP)
        assert res.user_facing_prop == HAS_PROP
        assert res.user_facing_value == "true"
        assert res.knora_type == KnoraValueType.BOOLEAN_VALUE
        assert len(res.value_metadata) == 1
        metadata = res.value_metadata.pop(0)
        assert metadata.property_type == TriplePropertyType.KNORA_COMMENT_ON_VALUE
        assert metadata.object_value == "comment"
        assert metadata.object_type == TripleObjectType.STRING

    def test_boolean_with_comment_empty_string(self):
        val = ParsedValue(HAS_PROP, "true", KnoraValueType.BOOLEAN_VALUE, None, "")
        res = _get_one_value(val, LIST_LOOKUP)
        assert res.user_facing_prop == HAS_PROP
        assert res.user_facing_value == "true"
        assert res.knora_type == KnoraValueType.BOOLEAN_VALUE
        assert len(res.value_metadata) == 1
        metadata = res.value_metadata.pop(0)
        assert metadata.property_type == TriplePropertyType.KNORA_COMMENT_ON_VALUE
        assert metadata.object_value == ""
        assert metadata.object_type == TripleObjectType.STRING

    def test_boolean_with_permissions_corr(self):
        val = ParsedValue(HAS_PROP, "true", KnoraValueType.BOOLEAN_VALUE, "public", None)
        res = _get_one_value(val, LIST_LOOKUP)
        assert res.user_facing_prop == HAS_PROP
        assert res.user_facing_value == "true"
        assert res.knora_type == KnoraValueType.BOOLEAN_VALUE
        assert len(res.value_metadata) == 1
        metadata = res.value_metadata.pop(0)
        assert metadata.property_type == TriplePropertyType.KNORA_PERMISSIONS
        assert metadata.object_value == "public"
        assert metadata.object_type == TripleObjectType.STRING

    def test_boolean_none(self):
        val = ParsedValue(HAS_PROP, None, KnoraValueType.BOOLEAN_VALUE, None, None)
        res = _get_one_value(val, LIST_LOOKUP)
        assert res.user_facing_prop == HAS_PROP
        assert res.user_facing_value == None  # noqa: E711 Comparison to `None`
        assert res.knora_type == KnoraValueType.BOOLEAN_VALUE
        assert not res.value_metadata

    def test_color_corr(self):
        val = ParsedValue(HAS_PROP, "#5d1f1e", KnoraValueType.COLOR_VALUE, None, None)
        res = _get_one_value(val, LIST_LOOKUP)
        assert res.user_facing_prop == HAS_PROP
        assert res.user_facing_value == "#5d1f1e"
        assert res.knora_type == KnoraValueType.COLOR_VALUE
        assert not res.value_metadata

    def test_date_corr_with_date_range_yyyy(self):
        val = ParsedValue(HAS_PROP, "CE:1849:CE:1850", KnoraValueType.DATE_VALUE, None, None)
        res = _get_one_value(val, LIST_LOOKUP)
        assert res.user_facing_prop == HAS_PROP
        assert res.user_facing_value == "CE:1849:CE:1850"
        assert res.knora_type == KnoraValueType.DATE_VALUE
        assert len(res.value_metadata) == 2
        start = next(x for x in res.value_metadata if x.property_type == TriplePropertyType.KNORA_DATE_START)
        assert start.object_value == "1849-01-01"
        assert start.object_type == TripleObjectType.DATE_YYYY_MM_DD
        end = next(x for x in res.value_metadata if x.property_type == TriplePropertyType.KNORA_DATE_END)
        assert end.object_value == "1850-01-01"
        assert end.object_type == TripleObjectType.DATE_YYYY_MM_DD

    def test_get_xsd_like_dates_only_start(self):
        date_str = "GREGORIAN:CE:1800-01-01"
        result = _get_xsd_like_dates(date_str)
        assert len(result) == 2
        start, end = result
        assert start.property_type == TriplePropertyType.KNORA_DATE_START
        assert start.object_value == "1800-01-01"
        assert start.object_type == TripleObjectType.DATE_YYYY_MM_DD
        assert end.property_type == TriplePropertyType.KNORA_DATE_END
        assert end.object_value == "1800-01-01"
        assert end.object_type == TripleObjectType.DATE_YYYY_MM_DD

    def test_get_xsd_like_dates_mixed_precision(self):
        date_str = "GREGORIAN:CE:1800-01-01:CE:1900"
        result = _get_xsd_like_dates(date_str)
        assert len(result) == 2
        start, end = result
        assert start.property_type == TriplePropertyType.KNORA_DATE_START
        assert start.object_value == "1800-01-01"
        assert start.object_type == TripleObjectType.DATE_YYYY_MM_DD
        assert end.property_type == TriplePropertyType.KNORA_DATE_END
        assert end.object_value == "1900-01-01"
        assert end.object_type == TripleObjectType.DATE_YYYY_MM_DD

    def test_get_xsd_like_dates_second_era_wrong(self):
        date_str = "GREGORIAN:CE:2000:BCE:1900"
        result = _get_xsd_like_dates(date_str)
        assert len(result) == 1
        start = result.pop(0)
        assert start.property_type == TriplePropertyType.KNORA_DATE_START
        assert start.object_value == "2000-01-01"
        assert start.object_type == TripleObjectType.DATE_YYYY_MM_DD

    @pytest.mark.parametrize(
        "date",
        [
            "BCE:2020-01-01:BCE:2021-02-02",  # BCE is not supported
            "BC:2020-01-01:BC:2021-02-02",  # BC is not supported
        ],
    )
    def test_dates_not_supported(self, date):
        result = _get_xsd_like_dates(date)
        assert not result

    @pytest.mark.parametrize(
        ("date", "expected_str"),
        [
            (SingleDate(era=None, year=1900, month=10, day=1), "1900-10-01"),
            (SingleDate(era=None, year=1900, month=10, day=None), "1900-10-01"),
            (SingleDate(era=None, year=900, month=None, day=None), "0900-01-01"),
            (SingleDate(era=Era.CE, year=900, month=None, day=None), "0900-01-01"),
            (SingleDate(era=Era.AD, year=900, month=None, day=None), "0900-01-01"),
        ],
    )
    def test_get_date_str(self, date, expected_str):
        result_str = _get_date_str(date)
        assert result_str == expected_str

    def test_decimal_corr(self):
        val = ParsedValue(HAS_PROP, "1.4", KnoraValueType.DECIMAL_VALUE, None, None)
        res = _get_one_value(val, LIST_LOOKUP)
        assert res.user_facing_prop == HAS_PROP
        assert res.user_facing_value == "1.4"
        assert res.knora_type == KnoraValueType.DECIMAL_VALUE
        assert not res.value_metadata

    def test_geoname_corr(self):
        val = ParsedValue(HAS_PROP, "1111111", KnoraValueType.GEONAME_VALUE, None, None)
        res = _get_one_value(val, LIST_LOOKUP)
        assert res.user_facing_prop == HAS_PROP
        assert res.user_facing_value == "1111111"
        assert res.knora_type == KnoraValueType.GEONAME_VALUE
        assert not res.value_metadata

    def test_geom_corr(self):
        geometry = """{
                    "status": "active",
                    "type": "polygon",
                    "lineWidth": 5,
                    "points": [{"x": 0.4, "y": 0.6},
                    {"x": 0.5, "y": 0.9},
                    {"x": 0.8, "y": 0.9},
                    {"x": 0.7, "y": 0.6}]
                    }"""
        val = ParsedValue(f"{KNORA_API_STR}hasGeometry", geometry, KnoraValueType.GEOM_VALUE, None, None)
        res = _get_one_value(val, LIST_LOOKUP)
        assert res.user_facing_prop == f"{KNORA_API_STR}hasGeometry"
        assert res.user_facing_value is not None
        assert res.knora_type == KnoraValueType.GEOM_VALUE
        assert not res.value_metadata

    def test_geom_wrong(self):
        val = ParsedValue(f"{KNORA_API_STR}hasGeometry", "invalid", KnoraValueType.GEOM_VALUE, None, None)
        res = _get_one_value(val, LIST_LOOKUP)
        assert res.user_facing_prop == f"{KNORA_API_STR}hasGeometry"
        assert res.user_facing_value == None  # noqa: E711 Comparison to `None`
        assert res.knora_type == KnoraValueType.GEOM_VALUE
        assert not res.value_metadata

    def test_geom_none(self):
        val = ParsedValue(f"{KNORA_API_STR}hasGeometry", None, KnoraValueType.GEOM_VALUE, None, None)
        res = _get_one_value(val, LIST_LOOKUP)
        assert res.user_facing_prop == f"{KNORA_API_STR}hasGeometry"
        assert res.user_facing_value == None  # noqa: E711 Comparison to `None`
        assert res.knora_type == KnoraValueType.GEOM_VALUE
        assert not res.value_metadata

    def test_int_corr(self):
        val = ParsedValue(HAS_PROP, "1", KnoraValueType.INT_VALUE, None, None)
        res = _get_one_value(val, LIST_LOOKUP)
        assert res.user_facing_prop == HAS_PROP
        assert res.user_facing_value == "1"
        assert res.knora_type == KnoraValueType.INT_VALUE
        assert not res.value_metadata

    def test_interval_corr(self):
        val = ParsedValue(f"{KNORA_API_STR}hasSegmentBounds", ("1", "2"), KnoraValueType.INTERVAL_VALUE, None, None)
        res = _get_one_value(val, LIST_LOOKUP)
        assert res.user_facing_prop == "http://api.knora.org/ontology/knora-api/v2#hasSegmentBounds"
        assert res.user_facing_value == None  # noqa: E711 Comparison to `None`
        assert res.knora_type == KnoraValueType.INTERVAL_VALUE
        assert len(res.value_metadata) == 2
        interval_start = next(
            x for x in res.value_metadata if x.property_type == TriplePropertyType.KNORA_INTERVAL_START
        )
        assert interval_start.object_value == "1"
        assert interval_start.object_type == TripleObjectType.DECIMAL
        interval_end = next(x for x in res.value_metadata if x.property_type == TriplePropertyType.KNORA_INTERVAL_END)
        assert interval_end.object_value == "2"
        assert interval_end.object_type == TripleObjectType.DECIMAL

    def test_list_corr(self):
        val = ParsedValue(HAS_PROP, ("list", "node"), KnoraValueType.LIST_VALUE, None, None)
        res = _get_one_value(val, LIST_LOOKUP)
        assert res.user_facing_prop == HAS_PROP
        assert res.user_facing_value == "http://rdfh.ch/lists/9999/n1"
        assert res.knora_type == KnoraValueType.LIST_VALUE
        assert not res.value_metadata

    def test_list_none(self):
        val = ParsedValue(HAS_PROP, ("list", None), KnoraValueType.LIST_VALUE, None, None)
        res = _get_one_value(val, LIST_LOOKUP)
        assert res.user_facing_prop == HAS_PROP
        assert res.user_facing_value == "list"
        assert res.knora_type == KnoraValueType.LIST_VALUE
        assert not res.value_metadata

    def test_link_corr(self):
        val = ParsedValue(HAS_PROP, "other_id", KnoraValueType.LINK_VALUE, None, None)
        res = _get_one_value(val, LIST_LOOKUP)
        assert res.user_facing_prop == HAS_PROP
        assert res.user_facing_value == "other_id"
        assert res.knora_type == KnoraValueType.LINK_VALUE
        assert not res.value_metadata

    def test_simple_text_corr(self):
        val = ParsedValue(HAS_PROP, "text", KnoraValueType.SIMPLETEXT_VALUE, None, None)
        res = _get_one_value(val, LIST_LOOKUP)
        assert res.user_facing_prop == HAS_PROP
        assert res.user_facing_value == "text"
        assert res.knora_type == KnoraValueType.SIMPLETEXT_VALUE
        assert not res.value_metadata

    def test_simple_text_wrong(self):
        val = ParsedValue(HAS_PROP, None, KnoraValueType.SIMPLETEXT_VALUE, None, None)
        res = _get_one_value(val, LIST_LOOKUP)
        assert res.user_facing_prop == HAS_PROP
        assert res.user_facing_value == None  # noqa: E711 Comparison to `None`
        assert res.knora_type == KnoraValueType.SIMPLETEXT_VALUE
        assert not res.value_metadata

    def test_richtext_corr(self):
        val = ParsedValue(HAS_PROP, "<p>Text</p>", KnoraValueType.RICHTEXT_VALUE, None, None)
        res = _get_one_value(val, LIST_LOOKUP)
        assert res.user_facing_prop == HAS_PROP
        assert res.user_facing_value == "<p>Text</p>"
        assert res.knora_type == KnoraValueType.RICHTEXT_VALUE
        assert not res.value_metadata

    def test_richtext_with_standoff(self, richtext_with_standoff):
        res = _get_one_value(richtext_with_standoff, LIST_LOOKUP)
        assert res.user_facing_prop == HAS_PROP
        assert res.user_facing_value == 'With <a class="salsah-link" href="IRI:link:IRI">link text</a>.'
        assert res.knora_type == KnoraValueType.RICHTEXT_VALUE
        assert not res.value_metadata

    def test_time_corr(self):
        val = ParsedValue(HAS_PROP, "2019-10-23T13:45:12.01-14:00", KnoraValueType.TIME_VALUE, None, None)
        res = _get_one_value(val, LIST_LOOKUP)
        assert res.user_facing_prop == HAS_PROP
        assert res.user_facing_value == "2019-10-23T13:45:12.01-14:00"
        assert res.knora_type == KnoraValueType.TIME_VALUE
        assert not res.value_metadata

    def test_uri_corr(self):
        val = ParsedValue(HAS_PROP, "https://dasch.swiss", KnoraValueType.URI_VALUE, None, None)
        res = _get_one_value(val, LIST_LOOKUP)
        assert res.user_facing_prop == HAS_PROP
        assert res.user_facing_value == "https://dasch.swiss"
        assert res.knora_type == KnoraValueType.URI_VALUE
        assert not res.value_metadata


class TestFileValue:
    def test_iiif(self):
        metadata = ParsedFileValueMetadata(None, None, None, None)
        val = ParsedFileValue("https://this/is/a/uri.jpg", KnoraValueType.STILL_IMAGE_IIIF, metadata)
        res = _get_file_value(val, AUTHORSHIP_LOOKUP)
        assert isinstance(res, RdfLikeValue)
        assert res.user_facing_prop == f"{KNORA_API_STR}hasStillImageFileValue"
        assert res.user_facing_value == "https://this/is/a/uri.jpg"
        assert res.knora_type == KnoraValueType.STILL_IMAGE_IIIF
        assert not res.value_metadata

    def test_iiif_with_legal_info(self):
        metadata = ParsedFileValueMetadata("http://rdfh.ch/licenses/cc-by-nc-4.0", "copy", "auth_id", None)
        result = _get_file_metadata(metadata, AUTHORSHIP_LOOKUP)
        assert len(result) == 3
        license_res = next(x for x in result if x.property_type == TriplePropertyType.KNORA_LICENSE)
        assert license_res.object_value == "http://rdfh.ch/licenses/cc-by-nc-4.0"
        assert license_res.object_type == TripleObjectType.IRI
        author_res = next(x for x in result if x.property_type == TriplePropertyType.KNORA_AUTHORSHIP)
        assert author_res.object_value == "Author Resolved"
        assert author_res.object_type == TripleObjectType.STRING
        copyright_res = next(x for x in result if x.property_type == TriplePropertyType.KNORA_COPYRIGHT_HOLDER)
        assert copyright_res.object_value == "copy"
        assert copyright_res.object_type == TripleObjectType.STRING

    def test_iiif_with_legal_info_unknown_authorship(self):
        metadata = ParsedFileValueMetadata("http://rdfh.ch/licenses/cc-by-nc-4.0", "copy", "unknonw", None)
        result = _get_file_metadata(metadata, AUTHORSHIP_LOOKUP)
        assert len(result) == 3
        license_res = next(x for x in result if x.property_type == TriplePropertyType.KNORA_LICENSE)
        assert license_res.object_value == "http://rdfh.ch/licenses/cc-by-nc-4.0"
        assert license_res.object_type == TripleObjectType.IRI
        author_res = next(x for x in result if x.property_type == TriplePropertyType.KNORA_AUTHORSHIP)
        assert author_res.object_value == ""
        assert author_res.object_type == TripleObjectType.STRING
        copyright_res = next(x for x in result if x.property_type == TriplePropertyType.KNORA_COPYRIGHT_HOLDER)
        assert copyright_res.object_value == "copy"
        assert copyright_res.object_type == TripleObjectType.STRING

    def test_no_file(self):
        metadata = ParsedFileValueMetadata(None, None, None, None)
        val = ParsedFileValue(None, None, metadata)
        assert not _get_file_value(val, AUTHORSHIP_LOOKUP)

    def test_unknown_extension(self):
        metadata = ParsedFileValueMetadata(None, None, None, None)
        val = ParsedFileValue("unknown.extension", None, metadata)
        assert not _get_file_value(val, AUTHORSHIP_LOOKUP)


@pytest.mark.parametrize(
    ("input_val", "expected"),
    [
        (("list", "node"), "http://rdfh.ch/lists/9999/n1"),
        (("list", ""), "list / "),
        (("list", None), "list"),
        ((None, "node"), "node"),
        (("", ""), " / "),
    ],
)
def test_get_list_value_str(input_val, expected):
    assert _get_list_value_str(input_val, LIST_LOOKUP) == expected


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

from pathlib import Path
from typing import Any
from unittest import mock

import numpy as np
import pandas as pd
import regex

from dsp_tools.commands.excel2json.models.input_error import DuplicateSheetProblem
from dsp_tools.commands.excel2json.models.input_error import DuplicatesInColumnProblem
from dsp_tools.commands.excel2json.models.input_error import ExcelFileProblem
from dsp_tools.commands.excel2json.models.input_error import InvalidSheetNameProblem
from dsp_tools.commands.excel2json.models.input_error import PositionInExcel
from dsp_tools.commands.excel2json.models.input_error import RequiredColumnMissingProblem
from dsp_tools.commands.excel2json.models.ontology import LanguageDict
from dsp_tools.error.exceptions import InputError
from dsp_tools.error.exceptions import UserFilepathNotFoundError

languages = ["en", "de", "fr", "it", "rm"]


def read_and_clean_all_sheets(excelfile: str | Path) -> dict[str, pd.DataFrame]:
    """
    This function reads an Excel file with all its sheets.
    If there is a ValueError, it patches the openpyxl part that causes the error
    and opens it with that patch.
    It cleans the dataframes and then returns them in the form {sheet_name: dataframe}.

    Args:
        excelfile: path to the Excel file

    Returns:
        All sheets of the excel file, in the form of a dictionary {sheet_name: dataframe}

    Raises:
        InputError: If the sheets are not correctly named
    """
    if not Path(excelfile).exists():
        raise UserFilepathNotFoundError(excelfile)
    try:
        df_dict = pd.read_excel(excelfile, sheet_name=None)
    except ValueError:
        # Pandas relies on openpyxl to parse XLSX files.
        # A strange behavior of openpyxl prevents pandas from opening files with some formatting properties
        # (unclear which formatting properties exactly).
        # Apparently, the excel2json test files have one of the unsupported formatting properties.
        # Credits: https://stackoverflow.com/a/70537454/14414188
        with mock.patch("openpyxl.styles.fonts.Font.family.max", new=100):
            df_dict = pd.read_excel(excelfile, sheet_name=None)
    _find_duplicate_col_names(str(excelfile), list(df_dict))
    try:
        return {name.strip(""): clean_data_frame(df) for name, df in df_dict.items()}
    except AttributeError:
        msg = InvalidSheetNameProblem(str(excelfile), list(df_dict.keys())).execute_error_protocol()
        raise InputError(msg) from None


def _find_duplicate_col_names(excelfile: str, col_names: list[str]) -> None:
    sheet_names = [str(x).lower().strip() for x in col_names]
    duplicate_names = list({x for x in sheet_names if sheet_names.count(x) > 1})
    if duplicate_names:
        msg = ExcelFileProblem(str(excelfile), [DuplicateSheetProblem(duplicate_names)]).execute_error_protocol()
        raise InputError(msg) from None


def clean_data_frame(df: pd.DataFrame) -> pd.DataFrame:
    """
    This function takes a pd.DataFrame and removes:
        - Leading and trailing spaces from the column names
        - Leading and trailing spaces from each cell and any characters in the cells that are not part of any known
        language, for example, linebreaks and replaces it with a pd.NA.
        - Removes all rows that are empty in all columns

    Args:
        df: The pd.DataFrame that is to be cleaned

    Returns:
        pd.DataFrame which has the above-mentioned removed
    """
    # Remove leading and trailing blanks in column names and make them lower case
    df = df.rename(columns=lambda x: x.strip().lower())
    # Remove the values of all cells that do not at least contain one character of any known language and removes
    # leading and trailing spaces.
    df = df.map(
        lambda x: str(x).strip() if pd.notna(x) and regex.search(r"[\w\p{L}]", str(x), flags=regex.U) else pd.NA
    )
    # drop all the rows that are entirely empty
    df = df.dropna(axis=0, how="all")
    return df


def check_contains_required_columns(
    df: pd.DataFrame, required_columns: set[str]
) -> None | RequiredColumnMissingProblem:
    """
    This function checks if all the columns from the set are in the pd.DataFrame.
    Additional columns to the ones in the set are allowed.

    Args:
        df: pd.DataFrame that is checked
        required_columns: set of column names

    Returns:
        An object if there is a problem else None.
    """
    if not required_columns.issubset(set(df.columns)):
        required = list(required_columns.difference(set(df.columns)))
        return RequiredColumnMissingProblem(columns=required)
    return None


def check_column_for_duplicate(df: pd.DataFrame, to_check_column: str) -> None | DuplicatesInColumnProblem:
    """
    This function checks if a specified column contains duplicate values.
    Empty cells (pd.NA) also count as duplicates.

    Args:
        df: pd.DataFrame that is checked for duplicates
        to_check_column: Name of the column that must not contain duplicates

    Returns:
        If there are problems it returns an object that stores the relevant user information.

    """
    if df[to_check_column].duplicated().any():
        duplicate_values = df[to_check_column][df[to_check_column].duplicated()].tolist()
        return DuplicatesInColumnProblem(
            column=to_check_column,
            duplicate_values=duplicate_values,
        )
    else:
        return None


def find_missing_required_values(
    df: pd.DataFrame, required_values_columns: list[str], sheetname: str | None = None
) -> list[PositionInExcel] | None:
    """
    If there are empty cells in the specified columns,
    It specifies the column and row numbers of all missing values and
    returns all the locations wrapped in the MissingValuesProblem

    Args:
        df: pd.DataFrame that is checked
        required_values_columns: a list of column names that may not contain empty cells
        sheetname: optional name of the Excel sheet

    Returns:
        Locations of missing values
        None if all are filled
    """
    if missing_values := check_required_values(df, required_values_columns):
        locations = []
        row_nums = get_wrong_row_numbers(missing_values)
        for col, nums in row_nums.items():
            locations.extend([PositionInExcel(sheet=sheetname, column=col, row=x) for x in nums])
        return locations
    return None


def check_required_values(df: pd.DataFrame, required_values_columns: list[str]) -> dict[str, pd.Series[bool]]:
    """
    If there are any empty cells in the column, it adds the column name and a boolean pd.Series to the dictionary.
    If there are no empty cells, then it is not included in the dictionary.
    If no column has any empty cells, then it returns an empty dictionary.

    Args:
        df: pd.DataFrame that is checked
        required_values_columns: a list of column names that may not contain empty cells

    Returns:
        a dictionary with the column names as key and pd.Series as values if there are any empty cells
    """
    # It checks if any of the values in a specified column are empty. If they are, they are added to the dictionary
    # with the column name as key and a boolean series as value that contain true for every pd.NA
    # If all the columns are filled, then it returns an empty dictionary.
    return {col: df[col].isnull() for col in required_values_columns if df[col].isnull().any()}


def _turn_bool_array_into_index_numbers(series: pd.Series[bool], true_remains: bool = True) -> list[int]:
    """
    This function takes a pd.Series containing boolean values.
    By default, this method extracts the index numbers of the True values.
    If the index numbers of the False values are required, the parameter "true_remains" should be turned to False.

    Args:
        series: pd.Series, which only contains True and False values
        true_remains: True if the index numbers of True are required, likewise with False

    Returns:
        A list of index numbers
    """
    # If the False are required, we need to invert the array.
    if not true_remains:
        series = ~series
    return list(series[series].index)


def get_wrong_row_numbers(
    wrong_row_dict: dict[str, pd.Series[bool]], true_remains: bool = True
) -> dict[str, list[int]]:
    """
    From the boolean pd.Series the index numbers of the True values are extracted.
    The resulting list is the new value of the dictionary.
    This new dictionary is taken and to each index number 2 is added, so that it corresponds to the Excel row number.

    Args:
        wrong_row_dict: The dictionary which contains column names and a boolean pd.Series
        true_remains: If True then the index of True is taken, if False then the index of False values is taken

    Returns:
        Dictionary with the column name as key and the row number as a list.
    """
    wrong_row_int_dict = {
        k: _turn_bool_array_into_index_numbers(series=v, true_remains=true_remains) for k, v in wrong_row_dict.items()
    }
    return {k: [x + 2 for x in v] for k, v in wrong_row_int_dict.items()}


def get_labels(df_row: pd.Series[Any]) -> LanguageDict:
    """
    This function takes a pd.Series which has "label_[language tag]" in the index.
    If the value of the index is not pd.NA, the language tag and the value are added to a dictionary.
    If it is empty, it is omitted from the dictionary.

    Args:
        df_row: a pd.Series (usually a row of a pd.DataFrame) from which the content of the columns containing the
                label is extracted

    Returns:
        A dictionary with the language tag and the content of the cell
    """
    labels = {lang: label for lang in languages if not pd.isna(label := df_row[f"label_{lang}"])}
    if not labels:
        labels = {lang: label for lang in languages if not pd.isna(label := df_row[lang])}
    return LanguageDict(labels)


def get_comments(df_row: pd.Series[Any]) -> LanguageDict | None:
    """
    This function takes a pd.Series which has "comment_[language tag]" in the index.
    If the value of the index is not pd.NA, the language tag and the value are added to a dictionary.
    If it is empty, it is omitted from the dictionary.

    Args:
        df_row: a pd.Series (usually a row of a pd.DataFrame) from which the content of the columns containing the
                comment is extracted

    Returns:
        A dictionary with the language tag and the content of the cell
    """
    comments = {lang: comment for lang in languages if not pd.isna(comment := df_row[f"comment_{lang}"])}
    return LanguageDict(comments) if comments else None


def find_one_full_cell_in_cols(df: pd.DataFrame, required_columns: list[str]) -> pd.Series[bool] | None:
    """
    This function takes a pd.DataFrame and a list of column names where at least one cell must have a value per row.
    A pd.Series with boolean values is returned, True if any rows do not have a value in at least one column

    Args:
        df: The pd.DataFrame which should be checked
        required_columns: A list of column names where at least one cell per row must have a value

    Returns:
        None if there is no problem or a pd.Series if there is a problem in a row
    """
    # The series has True if the cell is empty
    # In order to combine more than two arrays, we need to reduce the arrays, which takes a tuple
    result_arrays = tuple(df[col].isnull() for col in required_columns)
    # If all are True logical_and returns True otherwise False
    combined_array = np.logical_and.reduce(result_arrays)
    # if any of the values are True, it is turned into a pd.Series
    return pd.Series(combined_array) if any(combined_array) else None


def col_must_or_not_empty_based_on_other_col(
    df: pd.DataFrame,
    substring_list: list[str],
    substring_colname: str,
    check_empty_colname: str,
    must_have_value: bool,
) -> pd.Series[bool] | None:
    """
    It is presumed that the column "substring_colname" has no empty cells.
    Based on the string content of the individual rows, which is specified in the "substring_list",
    the cell in the column "check_empty_colname" is checked whether it is empty or not.
    The "substring_list" contains the different possibilities regarding the content of the cell.
    If the parameter "must_have_value" is True, then the cell in the "check_empty_colname" column must not be empty.
    If the parameter is set to False, then it must be empty.

    Args:
        df: The pd.DataFrame which is checked
        substring_list: A list of possible information that could be in the column "substring_colname"
        substring_colname: The name of the column that may contain any of the sub-strings
        check_empty_colname: The name of the column which is checked if it is empty or not
        must_have_value: True if the "check_empty_colname" should have a value or the reverse.

    Returns:
        None if all rows are correctly filled or empty.
        A series which contains True values for the rows, where it does
        not comply with the specifications.
    """
    na_series = df[check_empty_colname].isna()
    # If the cells have to be empty, we need to reverse the series
    if not must_have_value:
        na_series = ~na_series
    # This returns True if it finds the substring in the cell, they are joined in a RegEx "|" which denotes "or".
    # If it does not match any of the sub-strings, then the RegEx returns False,
    # which means that the value in the column "check_empty_colname" is not relevant.
    substring_array = df[substring_colname].str.contains("|".join(substring_list), na=False, regex=True)
    # If both are True logical_and returns True otherwise False
    combined_array = np.logical_and(na_series, substring_array)
    return pd.Series(combined_array) if any(combined_array) else None


def add_optional_columns(df: pd.DataFrame, optional_col_set: set[str]) -> pd.DataFrame:
    """
    Adds columns to a df if they are not already present.
    The content of the columns is empty.

    Args:
        df: Dataframe
        optional_col_set: set of columns that may be added

    Returns:
        Dataframe with additional columns if they were not present
    """
    in_df_cols = set(df.columns)
    if not optional_col_set.issubset(in_df_cols):
        additional_col = list(optional_col_set.difference(in_df_cols))
        additional_df = pd.DataFrame(columns=additional_col, index=df.index)
        df = pd.concat(objs=[df, additional_df], axis=1)
    return df



import json
from dataclasses import dataclass
from typing import Any
from typing import cast
from unittest.mock import patch

from dsp_tools.utils.request_utils import log_response


def test_log_response() -> None:
    response_mock = ResponseMock(
        status_code=200,
        headers={"Set-Cookie": "KnoraAuthenticationMFYGSLT", "Content-Type": "application/json"},
        text=json.dumps({"foo": "bar"}),
    )
    expected_output = {
        "status_code": 200,
        "headers": {"Set-Cookie": "***", "Content-Type": "application/json"},
        "content": {"foo": "bar"},
    }
    with patch("dsp_tools.utils.request_utils.logger.debug") as debug_mock:
        log_response(response_mock)  # type: ignore[arg-type]
        debug_mock.assert_called_once_with(f"RESPONSE: {json.dumps(expected_output)}")


@dataclass
class ResponseMock:
    status_code: int
    headers: dict[str, Any]
    text: str

    def json(self) -> dict[str, Any]:
        return cast(dict[str, Any], json.loads(self.text))



# mypy: disable-error-code="method-assign,no-untyped-def"
from unittest.mock import Mock
from unittest.mock import patch

import pytest

from dsp_tools.clients.legal_info_client_live import HTTP_LACKING_PERMISSIONS
from dsp_tools.clients.legal_info_client_live import LegalInfoClientLive
from dsp_tools.clients.legal_info_client_live import _is_last_page
from dsp_tools.error.exceptions import BadCredentialsError
from dsp_tools.error.exceptions import BaseError
from dsp_tools.utils.request_utils import RequestParameters

AUTH = Mock()
AUTH.get_token = Mock(return_value="tkn")


LICENSE_1 = {
    "id": "http://rdfh.ch/1",
    "uri": "https://uri.org/1",
    "labelEn": "label 1",
    "isRecommended": True,
    "isEnabled": True,
}

LICENSE_2 = {
    "id": "http://rdfh.ch/2",
    "uri": "https://uri.org/2",
    "labelEn": "label 2",
    "isRecommended": True,
    "isEnabled": True,
}

DATA_PAGE_1_OF_1 = {
    "data": [
        LICENSE_1,
    ],
    "pagination": {"pageSize": 1, "totalItems": 1, "totalPages": 1, "currentPage": 1},
}

PAGE_1_OF_2 = {"data": [LICENSE_1], "pagination": {"pageSize": 1, "totalItems": 2, "totalPages": 2, "currentPage": 1}}

PAGE_2_OF_2 = {"data": [LICENSE_2], "pagination": {"pageSize": 1, "totalItems": 2, "totalPages": 2, "currentPage": 2}}


class TestPostCopyrightHolders:
    @patch("dsp_tools.clients.legal_info_client_live.log_response")
    @patch("dsp_tools.clients.legal_info_client_live.log_request")
    def test_log_request(self, log_request: Mock, log_response: Mock):  # noqa: ARG002
        client = LegalInfoClientLive("http://api.com", "9999", AUTH)
        params = RequestParameters(
            method="POST",
            url="http://api.com/admin/projects/shortcode/9999/legal-info/copyright-holders",
            data={"data": ["1"]},
            timeout=60,
        )
        with patch("dsp_tools.clients.legal_info_client_live.requests.post") as post_mock:
            post_mock.return_value = Mock(status_code=200, ok=True)
            response = client._post_and_log_request("copyright-holders", ["1"])
        assert response.status_code == 200
        assert response.ok
        log_request_call_params = log_request.call_args_list[0].args[0]
        assert isinstance(log_request_call_params, RequestParameters)
        del log_request_call_params.headers
        assert log_request_call_params == params

    @patch("dsp_tools.clients.legal_info_client_live.log_response")
    @patch("dsp_tools.clients.legal_info_client_live.log_request")
    def test_post_ok(self, log_request: Mock, log_response: Mock):  # noqa: ARG002
        client = LegalInfoClientLive("http://api.com", "9999", AUTH)
        params = RequestParameters(
            method="POST",
            url="http://api.com/admin/projects/shortcode/9999/legal-info/copyright-holders",
            data={"data": ["1"]},
            timeout=60,
            headers={"Content-Type": "application/json", "Authorization": "Bearer tkn"},
        )
        with patch("dsp_tools.clients.legal_info_client_live.requests.post") as post_mock:
            post_mock.return_value = Mock(status_code=200, ok=True)
            client._post_and_log_request("copyright-holders", ["1"])
            post_mock.assert_called_once_with(
                url=params.url, headers=params.headers, data=params.data_serialized, timeout=params.timeout
            )

    def test_client_post_copyright_holders_lacking_permissions(self):
        client = LegalInfoClientLive("http://api.com", "9999", AUTH)
        expected_response = Mock(status_code=HTTP_LACKING_PERMISSIONS, ok=False)
        client._post_and_log_request = Mock(return_value=expected_response)
        with pytest.raises(BadCredentialsError):
            client.post_copyright_holders(["1"])

    def test_client_post_copyright_holders_unknown_status_code(self):
        client = LegalInfoClientLive("http://api.com", "9999", AUTH)
        expected_response = Mock(status_code=404, ok=False)
        client._post_and_log_request = Mock(return_value=expected_response)
        with pytest.raises(BaseError):
            client.post_copyright_holders(["1"])


class TestGetEnabledLicenses:
    def test_get_enabled_license_page(self):
        client = LegalInfoClientLive("http://api.com", "9999", AUTH)
        first_response = Mock(status_code=200, ok=True)
        first_response.json.return_value = PAGE_1_OF_2
        second_response = Mock(status_code=200, ok=True)
        second_response.json.return_value = PAGE_2_OF_2
        with patch.object(
            LegalInfoClientLive,
            attribute="_get_one_license_page",
            side_effect=[first_response, second_response],
        ):
            response = client.get_licenses_of_a_project()
            assert response == [LICENSE_1, LICENSE_2]

    def test_get_enabled_license_page_no_license(self):
        client = LegalInfoClientLive("http://api.com", "9999", AUTH)
        get_mock = Mock(
            status_code=200,
            ok=True,
        )
        get_mock.json.return_value = {
            "data": [],
            "pagination": {"pageSize": 0, "totalItems": 0, "totalPages": 1, "currentPage": 1},
        }
        with patch.object(
            LegalInfoClientLive,
            attribute="_get_one_license_page",
            side_effect=[get_mock],
        ):
            response = client.get_licenses_of_a_project(enabled_only=True)
            assert response == []

    @patch("dsp_tools.clients.legal_info_client_live.log_response")
    @patch("dsp_tools.clients.legal_info_client_live.log_request")
    def test_get_one_license_page(self, log_request: Mock, log_response: Mock):  # noqa: ARG002
        client = LegalInfoClientLive("http://api.com", "9999", AUTH)
        params = RequestParameters(
            method="GET",
            url="http://api.com/admin/projects/shortcode/9999/legal-info/licenses?page=1&page-size=25&order=Asc&showOnlyEnabled=true",
            timeout=60,
            headers={"Content-Type": "application/json", "Authorization": "Bearer tkn"},
        )
        with patch("dsp_tools.clients.legal_info_client_live.requests.get") as get_mock:
            mock_response = Mock(status_code=200, ok=True)
            mock_response.json.return_value = DATA_PAGE_1_OF_1
            get_mock.return_value = mock_response
            response = client._get_one_license_page(page_num=1, enabled_only=True)
            get_mock.assert_called_once_with(url=params.url, headers=params.headers, timeout=params.timeout)
        assert response.json() == DATA_PAGE_1_OF_1

    def test_insufficient_credentials(self):
        client = LegalInfoClientLive("http://api.com", "9999", AUTH)
        mock_response = Mock(status_code=HTTP_LACKING_PERMISSIONS, ok=False)
        mock_response.json.return_value = {}
        mock_response.headers = {}
        with patch("dsp_tools.clients.legal_info_client_live.requests.get") as get_mock:
            get_mock.return_value = mock_response
            with pytest.raises(BadCredentialsError):
                client._get_one_license_page(page_num=1, enabled_only=True)

    def test_unknown_status_code(self):
        client = LegalInfoClientLive("http://api.com", "9999", AUTH)
        mock_response = Mock(status_code=404, ok=False)
        mock_response.json.return_value = {}
        mock_response.headers = {}
        with patch("dsp_tools.clients.legal_info_client_live.requests.get") as get_mock:
            get_mock.return_value = mock_response
            with pytest.raises(BaseError):
                client._get_one_license_page(page_num=1, enabled_only=True)


def test_is_last_page_no():
    assert not _is_last_page(PAGE_1_OF_2)


def test_is_last_page_yes():
    assert _is_last_page(DATA_PAGE_1_OF_1)



from __future__ import annotations

from collections.abc import Collection
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Any

from dsp_tools.error.xmllib_warnings import MessageInfo
from dsp_tools.error.xmllib_warnings_util import emit_xmllib_input_type_mismatch_warning
from dsp_tools.error.xmllib_warnings_util import raise_xmllib_input_error
from dsp_tools.xmllib.internal.input_converters import check_and_fix_collection_input
from dsp_tools.xmllib.internal.input_converters import check_and_fix_is_non_empty_string
from dsp_tools.xmllib.models.config_options import NewlineReplacement
from dsp_tools.xmllib.models.config_options import Permissions
from dsp_tools.xmllib.models.internal.file_values import AbstractFileValue
from dsp_tools.xmllib.models.internal.file_values import FileValue
from dsp_tools.xmllib.models.internal.file_values import IIIFUri
from dsp_tools.xmllib.models.internal.file_values import Metadata
from dsp_tools.xmllib.models.internal.migration_metadata import MigrationMetadata
from dsp_tools.xmllib.models.internal.values import BooleanValue
from dsp_tools.xmllib.models.internal.values import ColorValue
from dsp_tools.xmllib.models.internal.values import DateValue
from dsp_tools.xmllib.models.internal.values import DecimalValue
from dsp_tools.xmllib.models.internal.values import GeonameValue
from dsp_tools.xmllib.models.internal.values import IntValue
from dsp_tools.xmllib.models.internal.values import LinkValue
from dsp_tools.xmllib.models.internal.values import ListValue
from dsp_tools.xmllib.models.internal.values import Richtext
from dsp_tools.xmllib.models.internal.values import SimpleText
from dsp_tools.xmllib.models.internal.values import TimeValue
from dsp_tools.xmllib.models.internal.values import UriValue
from dsp_tools.xmllib.models.internal.values import Value
from dsp_tools.xmllib.models.licenses.recommended import License
from dsp_tools.xmllib.value_checkers import is_nonempty_value
from dsp_tools.xmllib.value_checkers import is_valid_resource_id

LIST_SEPARATOR = "\n    - "


@dataclass
class Resource:
    res_id: str
    restype: str
    label: str
    values: list[Value] = field(default_factory=list)
    permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS
    file_value: AbstractFileValue | None = None
    migration_metadata: MigrationMetadata | None = None

    @staticmethod
    def create_new(
        res_id: str,
        restype: str,
        label: str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
    ) -> Resource:
        """
        Create a new resource.

        Args:
            res_id: resource ID
            restype: resource type
            label: resource label
            permissions: optional permissions of this resource

        Returns:
            Resource

        Examples:
            ```python
            resource = xmllib.Resource.create_new(
                res_id="ID",
                restype=":ResourceType",
                label="label",
            )
            ```

            ```python
            # resource with restricted permissions
            resource = xmllib.Resource.create_new(
                res_id="ID",
                restype=":ResourceType",
                label="label",
                permissions=xmllib.Permissions.PRIVATE,
            )
            ```
        """
        if not is_valid_resource_id(res_id):
            emit_xmllib_input_type_mismatch_warning(
                expected_type="xsd:ID", value=res_id, res_id=res_id, value_field="resource ID"
            )
        if not is_nonempty_value(restype):
            emit_xmllib_input_type_mismatch_warning(
                expected_type="resource type", value=restype, res_id=res_id, value_field="restype"
            )
        lbl = check_and_fix_is_non_empty_string(value=label, res_id=res_id, value_field="label")

        return Resource(
            res_id=res_id,
            restype=restype,
            label=lbl,
            permissions=permissions,
        )

    #######################
    # BooleanValue
    #######################

    def add_bool(
        self,
        prop_name: str,
        value: bool | str | int | float,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        Add a boolean value to the resource.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#boolean)

        Conversions:
         - The conversion is case-insensitive, meaning that the words can also be capitalised.
         - `false`, `0`, `0.0`, `no`, `non`, `nein` -> `false`
         - `true`, `1`, `1.0`, `yes`, `oui`, `ja` -> `true`

        Args:
            prop_name: name of the property
            value: value to add
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added value

        Examples:
            ```python
            resource = resource.add_bool(
                prop_name=":propName",
                value=True
            )
            ```
        """
        self.values.append(
            BooleanValue.new(
                value=value, prop_name=prop_name, permissions=permissions, comment=comment, resource_id=self.res_id
            )
        )
        return self

    def add_bool_optional(
        self,
        prop_name: str,
        value: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        If the value is not empty, add it to the resource, otherwise return the resource unchanged.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#boolean)

        Conversions:
         - The conversion is case-insensitive, meaning that the words can also be capitalised.
         - `false`, `0`, `0.0`, `no`, `non`, `nein` -> `false`
         - `true`, `1`, `1.0`, `yes`, `oui`, `ja` -> `true`

        Args:
            prop_name: name of the property
            value: value to add or empty value
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added value if it was not empty, else the unchanged original resource.

        Examples:
            ```python
            resource = resource.add_bool_optional(
                prop_name=":propName",
                value=True
            )
            ```

            ```python
            resource = resource.add_bool_optional(
                prop_name=":propName",
                value=None,
            )
            ```
        """
        if is_nonempty_value(value):
            self.add_bool(value=value, prop_name=prop_name, permissions=permissions, comment=comment)
        return self

    #######################
    # ColorValue
    #######################

    def add_color(
        self,
        prop_name: str,
        value: str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        Add a color value to the resource.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#color)

        Args:
            prop_name: name of the property
            value: value to add
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added value

        Examples:
            ```python
            resource = resource.add_color(
                prop_name=":propName",
                value="#00ff66",
            )
            ```
        """
        self.values.append(
            ColorValue.new(
                value=value, prop_name=prop_name, permissions=permissions, comment=comment, resource_id=self.res_id
            )
        )
        return self

    def add_color_multiple(
        self,
        prop_name: str,
        values: Collection[str],
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        Add several color values to the resource.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#color)

        Args:
            prop_name: name of the property
            values: values to add
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added values

        Examples:
            ```python
            resource = resource.add_color_multiple(
                prop_name=":propName",
                values=["#00ff66", "#00ff55"],
            )
            ```
        """
        vals = check_and_fix_collection_input(values, prop_name, self.res_id)
        for v in vals:
            self.add_color(prop_name, v, permissions, comment)
        return self

    def add_color_optional(
        self,
        prop_name: str,
        value: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        If the value is not empty, add it to the resource, otherwise return the resource unchanged.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#color)

        Args:
            prop_name: name of the property
            value: value to add or empty value
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added value if it was not empty, else the unchanged original resource.

        Examples:
            ```python
            resource = resource.add_color_optional(
                prop_name=":propName",
                value="#00ff66",
            )
            ```

            ```python
            resource = resource.add_color_optional(
                prop_name=":propName",
                value=None,
            )
            ```
        """
        if is_nonempty_value(value):
            self.add_color(prop_name, value, permissions, comment)
        return self

    #######################
    # DateValue
    #######################

    def add_date(
        self,
        prop_name: str,
        value: str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        Add a date value to the resource.
        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#date)

        Args:
            prop_name: name of the property
            value: value to add
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added value

        Examples:
            ```python
            resource = resource.add_date(
                prop_name=":propName",
                value="GREGORIAN:CE:2014-01-31:CE:2014-01-31",
            )
            ```
        """
        self.values.append(
            DateValue.new(
                value=value, prop_name=prop_name, permissions=permissions, comment=comment, resource_id=self.res_id
            )
        )
        return self

    def add_date_multiple(
        self,
        prop_name: str,
        values: Collection[str],
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        Add several date values to the resource.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#date)

        Args:
            prop_name: name of the property
            values: values to add
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added values

        Examples:
            ```python
            resource = resource.add_date_multiple(
                prop_name=":propName",
                values=["GREGORIAN:CE:2014-01-31:CE:2014-01-31", "CE:1849:CE:1850"],
            )
            ```
        """
        vals = check_and_fix_collection_input(values, prop_name, self.res_id)
        for v in vals:
            self.add_date(prop_name, v, permissions, comment)
        return self

    def add_date_optional(
        self,
        prop_name: str,
        value: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        If the value is not empty, add it to the resource, otherwise return the resource unchanged.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#date)

        Args:
            prop_name: name of the property
            value: value to add or empty value
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added value if it was not empty, else the unchanged original resource.

        Examples:
            ```python
            resource = resource.add_date_optional(
                prop_name=":propName",
                value="CE:1849:CE:1850",
            )
            ```

            ```python
            resource = resource.add_date_optional(
                prop_name=":propName",
                value=None,
            )
            ```
        """
        if is_nonempty_value(value):
            self.add_date(prop_name, value, permissions, comment)
        return self

    #######################
    # DecimalValue
    #######################

    def add_decimal(
        self,
        prop_name: str,
        value: float | int | str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        Add a decimal value to the resource.
        If the value is provided as string, it must be convertible to integer or float.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#decimal)

        Args:
            prop_name: name of the property
            value: value to add
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added value

        Examples:
            ```python
            resource = resource.add_decimal(
                prop_name=":propName",
                value=1.4,
            )
            ```
        """
        self.values.append(
            DecimalValue.new(
                value=value, prop_name=prop_name, permissions=permissions, comment=comment, resource_id=self.res_id
            )
        )
        return self

    def add_decimal_multiple(
        self,
        prop_name: str,
        values: Collection[float | int | str],
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        Add several decimal values to the resource.
        If the values are provided as string, they must be convertible to integer or float.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#decimal)

        Args:
            prop_name: name of the property
            values: values to add
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added values

        Examples:
            ```python
            resource = resource.add_decimal_multiple(
                prop_name=":propName",
                values=[1.4, 2.9],
            )
            ```
        """
        vals = check_and_fix_collection_input(values, prop_name, self.res_id)
        for v in vals:
            self.add_decimal(prop_name, v, permissions, comment)
        return self

    def add_decimal_optional(
        self,
        prop_name: str,
        value: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        If the value is not empty, add it to the resource, otherwise return the resource unchanged.
        If the value is provided as string, it must be convertible to integer or float.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#decimal)

        Args:
            prop_name: name of the property
            value: value to add or empty value
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added value if it was not empty, else the unchanged original resource.

        Examples:
            ```python
            resource = resource.add_decimal_optional(
                prop_name=":propName",
                value=1.0,
            )
            ```

            ```python
            resource = resource.add_decimal_optional(
                prop_name=":propName",
                value=None,
            )
            ```
        """
        if is_nonempty_value(value):
            self.add_decimal(prop_name, value, permissions, comment)
        return self

    #######################
    # GeonameValue
    #######################

    def add_geoname(
        self,
        prop_name: str,
        value: int | str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        Add a [geonames.org](https://www.geonames.org/) value to the resource.
        The [geonames.org](https://www.geonames.org/) identifier must be provided as integer
        or string that is convertible to integer.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#geoname)


        Args:
            prop_name: name of the property
            value: value to add
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added value

        Examples:
            ```python
            resource = resource.add_geoname(
                prop_name=":propName",
                value="2761369",
            )
            ```
        """
        self.values.append(
            GeonameValue.new(
                value=value, prop_name=prop_name, permissions=permissions, comment=comment, resource_id=self.res_id
            )
        )
        return self

    def add_geoname_multiple(
        self,
        prop_name: str,
        values: Collection[int | str],
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        Add several [geonames.org](https://www.geonames.org/) values to the resource.
        The [geonames.org](https://www.geonames.org/) identifiers must be provided as integers
        or strings that are convertible to integers.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#geoname)

        Args:
            prop_name: name of the property
            values: values to add
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added values

        Examples:
            ```python
            resource = resource.add_geoname_multiple(
                prop_name=":propName",
                values=["2761369", "2661604"],
            )
            ```
        """
        vals = check_and_fix_collection_input(values, prop_name, self.res_id)
        for v in vals:
            self.add_geoname(prop_name, v, permissions, comment)
        return self

    def add_geoname_optional(
        self,
        prop_name: str,
        value: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        If the value is not empty, add it to the resource, otherwise return the resource unchanged.
        The [geonames.org](https://www.geonames.org/) identifier must be provided as integer
        or string that is convertible to integer.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#geoname)

        Args:
            prop_name: name of the property
            value: value to add or empty value
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added value if it was not empty, else the unchanged original resource.

        Examples:
            ```python
            resource = resource.add_geoname_optional(
                prop_name=":propName",
                value="2661604",
            )
            ```

            ```python
            resource = resource.add_geoname_optional(
                prop_name=":propName",
                value=None,
            )
            ```
        """
        if is_nonempty_value(value):
            self.add_geoname(prop_name, value, permissions, comment)
        return self

    #######################
    # IntValue
    #######################

    def add_integer(
        self,
        prop_name: str,
        value: int | str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        Add an integer value to the resource.
        If the value is provided as string, it must be convertible to integer.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#integer)

        Args:
            prop_name: name of the property
            value: value to add
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added value

        Examples:
            ```python
            resource = resource.add_integer(
                prop_name=":propName",
                value=399,
            )
            ```
        """
        self.values.append(
            IntValue.new(
                value=value, prop_name=prop_name, permissions=permissions, comment=comment, resource_id=self.res_id
            )
        )
        return self

    def add_integer_multiple(
        self,
        prop_name: str,
        values: Collection[int | str],
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        Add several integer values to the resource.
        If the values are provided as string, they must be convertible to integer.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#integer)

        Args:
            prop_name: name of the property
            values: values to add
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added values

        Examples:
            ```python
            resource = resource.add_integer_multiple(
                prop_name=":propName",
                values=[24, "66"],
            )
            ```
        """
        vals = check_and_fix_collection_input(values, prop_name, self.res_id)
        for v in vals:
            self.add_integer(prop_name, v, permissions, comment)
        return self

    def add_integer_optional(
        self,
        prop_name: str,
        value: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        If the value is not empty, add it to the resource, otherwise return the resource unchanged.
        If the value is provided as string, it must be convertible to integer.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#integer)

        Args:
            prop_name: name of the property
            value: value to add or empty value
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added value if it was not empty, else the unchanged original resource.

        Examples:
            ```python
            resource = resource.add_integer_optional(
                prop_name=":propName",
                value="2604",
            )
            ```

            ```python
            resource = resource.add_integer_optional(
                prop_name=":propName",
                value=None,
            )
            ```
        """
        if is_nonempty_value(value):
            self.add_integer(prop_name, value, permissions, comment)
        return self

    #######################
    # LinkValue
    #######################

    def add_link(
        self,
        prop_name: str,
        value: str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        Add a link value to the resource, in the form of an ID of another resource.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#resptr)

        Args:
            prop_name: name of the property
            value: target resource ID
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added value

        Examples:
            ```python
            resource = resource.add_link(
                prop_name=":propName",
                value="target_resource_id",
            )
            ```
        """
        self.values.append(
            LinkValue.new(
                value=value, prop_name=prop_name, permissions=permissions, comment=comment, resource_id=self.res_id
            )
        )
        return self

    def add_link_multiple(
        self,
        prop_name: str,
        values: Collection[str],
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        Add several link values to the resource, in the form of IDs of other resources.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#resptr)

        Args:
            prop_name: name of the property
            values: list of target resources IDs
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added values

        Examples:
            ```python
            resource = resource.add_link_multiple(
                prop_name=":propName",
                values=["target_resource_id_1", "target_resource_id_2"],
            )
            ```
        """
        vals = check_and_fix_collection_input(values, prop_name, self.res_id)
        for v in vals:
            self.add_link(prop_name, v, permissions, comment)
        return self

    def add_link_optional(
        self,
        prop_name: str,
        value: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        If the value is not empty, add it to the resource, otherwise return the resource unchanged.
        If non-empty, the value must be an ID of another resource.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#resptr)

        Args:
            prop_name: name of the property
            value: target resource ID or empty value
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added value if it was not empty, else the unchanged original resource.

        Examples:
            ```python
            resource = resource.add_link_optional(
                prop_name=":propName",
                value="target_resource_id",
            )
            ```

            ```python
            resource = resource.add_link_optional(
                prop_name=":propName",
                value=None,
            )
            ```
        """
        if is_nonempty_value(value):
            self.add_link(prop_name, value, permissions, comment)
        return self

    #######################
    # ListValue
    #######################

    def add_list(
        self,
        prop_name: str,
        list_name: str | int | float,
        value: str | int | float,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        Add a list value to the resource, i.e. a name of a list node.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#list)

        Args:
            prop_name: name of the property
            list_name: name of the list (N.B. not the label, but the name of the list)
            value: name of a list node (N.B. not the label, but the name of the list node)
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added value

        Examples:
            ```python
            resource = resource.add_list(
                prop_name=":propName",
                list_name="listName",
                value="nodeName",
            )
            ```
        """
        self.values.append(
            ListValue.new(
                value=value,
                list_name=list_name,
                prop_name=prop_name,
                permissions=permissions,
                comment=comment,
                resource_id=self.res_id,
            )
        )
        return self

    def add_list_multiple(
        self,
        prop_name: str,
        list_name: str | int | float,
        values: Collection[str | int | float],
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        Add several list values to the resource, i.e. names of list nodes.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#list)

        Args:
            prop_name: name of the property
            list_name: name of the list (N.B. not the label, but the name of the list)
            values: names of list nodes (N.B. not the labels, but the names of the list nodes)
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added values

        Examples:
            ```python
            resource = resource.add_list_multiple(
                prop_name=":propName",
                list_name="listName",
                values=["nodeName_1", "nodeName_2"],
            )
            ```
        """
        vals = check_and_fix_collection_input(values, prop_name, self.res_id)
        for v in vals:
            self.add_list(prop_name, list_name, v, permissions, comment)
        return self

    def add_list_optional(
        self,
        prop_name: str,
        list_name: str | int | float,
        value: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        If the value is not empty, add it to the resource, otherwise return the resource unchanged.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#list)

        Args:
            prop_name: name of the property
            list_name: name of the list (N.B. not the label, but the name of the list)
            value: name of a list node (N.B. not the label, but the name of the list node) or empty value
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added value if it was not empty, else the unchanged original resource.

        Examples:
            ```python
            resource = resource.add_list_optional(
                prop_name=":propName",
                list_name="listName",
                value="nodeName",
            )
            ```

            ```python
            resource = resource.add_list_optional(
                prop_name=":propName",
                list_name="listName",
                value=None,
            )
            ```
        """
        if is_nonempty_value(value):
            self.add_list(prop_name, list_name, value, permissions, comment)
        return self

    #######################
    # TextValue: SimpleText
    #######################

    def add_simpletext(
        self,
        prop_name: str,
        value: str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        Add a simple text value to the resource.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#text)

        Args:
            prop_name: name of the property
            value: value to add
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added value

        Examples:
            ```python
            resource = resource.add_simpletext(
                prop_name=":propName",
                value="text",
            )
            ```
        """
        self.values.append(
            SimpleText.new(
                value=value, prop_name=prop_name, permissions=permissions, comment=comment, resource_id=self.res_id
            )
        )
        return self

    def add_simpletext_multiple(
        self,
        prop_name: str,
        values: Collection[str],
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        Add several simple text values to the resource.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#text)

        Args:
            prop_name: name of the property
            values: values to add
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added values

        Examples:
            ```python
            resource = resource.add_simpletext_multiple(
                prop_name=":propName",
                values=["text 1", "text 2"],
            )
            ```
        """
        vals = check_and_fix_collection_input(values, prop_name, self.res_id)
        for v in vals:
            self.add_simpletext(prop_name, v, permissions, comment)
        return self

    def add_simpletext_optional(
        self,
        prop_name: str,
        value: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        If the value is not empty, add it to the resource, otherwise return the resource unchanged.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#text)

        Args:
            prop_name: name of the property
            value: value to add or empty value
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added value if it was not empty, else the unchanged original resource.

        Examples:
            ```python
            resource = resource.add_simpletext_optional(
                prop_name=":propName",
                value="text",
            )
            ```

            ```python
            resource = resource.add_simpletext_optional(
                prop_name=":propName",
                value=None,
            )
            ```
        """
        if is_nonempty_value(value):
            self.add_simpletext(prop_name, value, permissions, comment)
        return self

    #######################
    # TextValue: Richtext
    #######################

    def add_richtext(
        self,
        prop_name: str,
        value: str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
        newline_replacement: NewlineReplacement = NewlineReplacement.LINEBREAK,
    ) -> Resource:
        """
        Add a rich text value to the resource.
        Rich text values must be provided as strings, potentially containing DSP Standard Standoff Markup XML tags
        as [documented here.](https://docs.dasch.swiss/latest/DSP-API/03-endpoints/api-v2/text/standard-standoff/)
        Only the documented tags are allowed.

        Conversions:
            By default, replace newline characters inside the text value with `<br/>`, which preserves the linebreak.
            Without this replacement, the newline would disappear, because `\\n` is meaningless in an XML file.
            [Click here for more details](https://docs.dasch.swiss/latest/DSP-TOOLS/xmllib-api-reference/value-converters/#xmllib.value_converters.replace_newlines_with_tags)

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#text)

        Args:
            prop_name: name of the property
            value: value to add
            permissions: optional permissions of this value
            comment: optional comment
            newline_replacement: options how to deal with `\\n` inside the text value. Default: replace with `<br/>`

        Returns:
            The original resource, with the added value

        Examples:
            ```python
            resource = resource.add_richtext(
                prop_name=":propName",
                value="line 1\\nline 2",
            )
            ```

            ```python
            # changing the replacement options for newlines `\\n`
            resource = resource.add_richtext(
                prop_name=":propName",
                value="line 1\\nline 2",
                newline_replacement=xmllib.NewlineReplacement.PARAGRAPH,
            )
            ```
        """
        self.values.append(
            Richtext.new(
                value=value,
                prop_name=prop_name,
                permissions=permissions,
                comment=comment,
                resource_id=self.res_id,
                newline_replacement=newline_replacement,
            )
        )
        return self

    def add_richtext_multiple(
        self,
        prop_name: str,
        values: Collection[str],
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
        newline_replacement: NewlineReplacement = NewlineReplacement.LINEBREAK,
    ) -> Resource:
        """
        Add several rich text values to the resource.
        Rich text values must be provided as strings, potentially containing DSP Standard Standoff Markup XML tags
        as [documented here.](https://docs.dasch.swiss/latest/DSP-API/03-endpoints/api-v2/text/standard-standoff/)
        Only the documented tags are allowed.

        Conversions:
            By default, replace newline characters inside the text value with `<br/>`, which preserves the linebreak.
            Without this replacement, the newline would disappear, because `\\n` is meaningless in an XML file.
            [Click here for more details](https://docs.dasch.swiss/latest/DSP-TOOLS/xmllib-api-reference/value-converters/#xmllib.value_converters.replace_newlines_with_tags)

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#text)

        Args:
            prop_name: name of the property
            values: values to add
            permissions: optional permissions of this value
            comment: optional comment
            newline_replacement: options how to deal with `\\n` inside the text value. Default: replace with `<br/>`

        Returns:
            The original resource, with the added values

        Examples:
            ```python
            resource = resource.add_richtext_multiple(
                prop_name=":propName",
                values=["<strong>Bold text</strong>", "text 2"],
            )
            ```
        """
        vals = check_and_fix_collection_input(values, prop_name, self.res_id)
        self.values.extend(
            [
                Richtext.new(
                    value=v,
                    prop_name=prop_name,
                    permissions=permissions,
                    comment=comment,
                    resource_id=self.res_id,
                    newline_replacement=newline_replacement,
                )
                for v in vals
            ]
        )
        return self

    def add_richtext_optional(
        self,
        prop_name: str,
        value: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
        newline_replacement: NewlineReplacement = NewlineReplacement.LINEBREAK,
    ) -> Resource:
        """
        If the value is not empty, add it to the resource, otherwise return the resource unchanged.
        Rich text values must be provided as strings, potentially containing DSP Standard Standoff Markup XML tags
        as [documented here.](https://docs.dasch.swiss/latest/DSP-API/03-endpoints/api-v2/text/standard-standoff/)
        Only the documented tags are allowed.

        Conversions:
            By default, replace newline characters inside the text value with `<br/>`, which preserves the linebreak.
            Without this replacement, the newline would disappear, because `\\n` is meaningless in an XML file.
            [Click here for more details](https://docs.dasch.swiss/latest/DSP-TOOLS/xmllib-api-reference/value-converters/#xmllib.value_converters.replace_newlines_with_tags)

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#text)

        Args:
            prop_name: name of the property
            value: value to add or empty value
            permissions: optional permissions of this value
            comment: optional comment
            newline_replacement: options how to deal with `\\n` inside the text value. Default: replace with `<br/>`

        Returns:
            The original resource, with the added value if it was not empty, else the unchanged original resource.

        Examples:
            ```python
            resource = resource.add_richtext_optional(
                prop_name=":propName",
                value="<strong>Bold text</strong>",
            )
            ```

            ```python
            resource = resource.add_richtext_optional(
                prop_name=":propName",
                value=None,
            )
            ```
        """
        if is_nonempty_value(value):
            return self.add_richtext(prop_name, value, permissions, comment, newline_replacement)
        return self

    #######################
    # TimeValue
    #######################

    def add_time(
        self,
        prop_name: str,
        value: str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        Add a time value to the resource.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#time)

        Args:
            prop_name: name of the property
            value: value to add
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added value

        Examples:
            ```python
            resource = resource.add_time(
                prop_name=":propName",
                value="2019-10-23T13:45:12Z",
            )
            ```
        """
        self.values.append(
            TimeValue.new(
                value=value, prop_name=prop_name, permissions=permissions, comment=comment, resource_id=self.res_id
            )
        )
        return self

    def add_time_multiple(
        self,
        prop_name: str,
        values: Collection[str],
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        Add several time values to the resource.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#time)

        Args:
            prop_name: name of the property
            values: values to add
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added values

        Examples:
            ```python
            resource = resource.add_time_multiple(
                prop_name=":propName",
                values=["2019-10-23T13:45:12Z", "2009-10-10T12:00:00-05:00"],
            )
            ```
        """
        vals = check_and_fix_collection_input(values, prop_name, self.res_id)
        for v in vals:
            self.add_time(prop_name, v, permissions, comment)
        return self

    def add_time_optional(
        self,
        prop_name: str,
        value: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        If the value is not empty, add it to the resource, otherwise return the resource unchanged.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#time)

        Args:
            prop_name: name of the property
            value: value to add or empty value
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added value if it was not empty, else the unchanged original resource.

        Examples:
            ```python
            resource = resource.add_time_optional(
                prop_name=":propName",
                value="2019-10-23T13:45:12Z",
            )
            ```

            ```python
            resource = resource.add_time_optional(
                prop_name=":propName",
                value=None,
            )
            ```
        """
        if is_nonempty_value(value):
            self.add_time(prop_name, value, permissions, comment)
        return self

    #######################
    # UriValue
    #######################

    def add_uri(
        self,
        prop_name: str,
        value: str,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        Add a URI value to the resource.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#uri)

        Args:
            prop_name: name of the property
            value: value to add
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added value

        Examples:
            ```python
            resource = resource.add_uri(
                prop_name=":propName",
                value="https://dasch.swiss",
            )
            ```
        """
        self.values.append(
            UriValue.new(
                value=value, prop_name=prop_name, permissions=permissions, comment=comment, resource_id=self.res_id
            )
        )
        return self

    def add_uri_multiple(
        self,
        prop_name: str,
        values: Collection[str],
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        Add several URI values to the resource.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#uri)

        Args:
            prop_name: name of the property
            values: values to add
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added values

        Examples:
            ```python
            resource = resource.add_uri_multiple(
                prop_name=":propName",
                values=["https://dasch.swiss", "https://docs.dasch.swiss"],
            )
            ```
        """
        vals = check_and_fix_collection_input(values, prop_name, self.res_id)
        for v in vals:
            self.add_uri(prop_name, v, permissions, comment)
        return self

    def add_uri_optional(
        self,
        prop_name: str,
        value: Any,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        If the value is not empty, add it to the resource, otherwise return the resource unchanged.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#uri)

        Args:
            prop_name: name of the property
            value: value to add or empty value
            permissions: optional permissions of this value
            comment: optional comment

        Returns:
            The original resource, with the added value if it was not empty, else the unchanged original resource.

        Examples:
            ```python
            resource = resource.add_uri_optional(
                prop_name=":propName",
                value="https://dasch.swiss",
            )
            ```

            ```python
            resource = resource.add_uri_optional(
                prop_name=":propName",
                value=None,
            )
            ```
        """
        if is_nonempty_value(value):
            self.add_uri(prop_name, value, permissions, comment)
        return self

    #######################
    # AbstractFileValue
    #######################

    def add_file(
        self,
        filename: str | Path,
        license: License | None = None,
        copyright_holder: str | None = None,
        authorship: list[str] | None = None,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        Add a file (bitstream) to the resource.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#bitstream)

        Args:
            filename: path to the file
            license: License of the file (predefined or custom) [see the documentation for the options.](https://docs.dasch.swiss/latest/DSP-TOOLS/xmllib-api-reference/licenses/recommended/).
                A license states the circumstances how you are allowed to share/reuse something.
            copyright_holder: The person or institution who owns the economic rights of something.
            authorship: The (natural) person who authored something.
            permissions: optional permissions of this file
            comment: optional comment

        Raises:
            XmllibInputError: If the resource already has a file or IIIF URI value

        Returns:
            The original resource, with the added value

        Examples:
            ```python
            resource = resource.add_file(
                filename="images/cat.jpg",
                license=xmllib.LicenseRecommended.DSP.PUBLIC_DOMAIN,
                copyright_holder="Meow University",
                authorship=["Kitty Meow"],
            )
            ```

            ```python
            # a file with restricted view permissions
            resource = resource.add_file(
                filename="images/dog.jpg",
                license=xmllib.LicenseRecommended.CC.BY_NC_ND,
                copyright_holder="Bark University",
                authorship=["Bark McDog"],
                permissions=xmllib.Permissions.LIMITED_VIEW,
            )
            ```
        """
        if self.file_value:
            msg_info = MessageInfo(
                f"This resource already contains a file with the name: '{self.file_value.value}'. "
                f"The new file with the name '{filename}' cannot be added.",
                resource_id=self.res_id,
                field="file / iiif-uri",
            )
            raise_xmllib_input_error(msg_info)
        meta = Metadata.new(
            license=license,
            copyright_holder=copyright_holder,
            authorship=authorship,
            permissions=permissions,
            resource_id=self.res_id,
        )
        self.file_value = FileValue.new(value=filename, metadata=meta, comment=comment, resource_id=self.res_id)
        return self

    def add_iiif_uri(
        self,
        iiif_uri: str,
        license: License | None = None,
        copyright_holder: str | None = None,
        authorship: list[str] | None = None,
        permissions: Permissions = Permissions.PROJECT_SPECIFIC_PERMISSIONS,
        comment: str | None = None,
    ) -> Resource:
        """
        Add a IIIF URI to the resource.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#iiif-uri)

        Args:
            iiif_uri: valid IIIF URI
            license: License of the file (predefined or custom) [see the documentation for the options.](https://docs.dasch.swiss/latest/DSP-TOOLS/xmllib-api-reference/licenses/recommended/).
                A license states the circumstances how you are allowed to share/reuse something.
            copyright_holder: The person or institution who owns the economic rights of something.
            authorship: The (natural) person who authored something.
            permissions: optional permissions of this file
            comment: optional comment

        Raises:
            XmllibInputError: If the resource already has a file or IIIF URI value

        Returns:
            The original resource, with the added value

        Examples:
            ```python
            resource = resource.add_iiif_uri(
                iiif_uri="https://iiif.wellcomecollection.org/image/b20432033_B0008608.JP2/full/1338%2C/0/default.jpg",
                license=xmllib.LicenseRecommended.CC.BY_NC,
                copyright_holder="Wellcome Collection",
                authorship=["Cavanagh, Annie"]
            )
            ```
        """
        if self.file_value:
            msg_info = MessageInfo(
                f"This resource already contains a file with the name: '{self.file_value.value}'. "
                f"The new file with the name '{iiif_uri}' cannot be added.",
                resource_id=self.res_id,
                field="file / iiif-uri",
            )
            raise_xmllib_input_error(msg_info)
        meta = Metadata.new(
            license=license,
            copyright_holder=copyright_holder,
            authorship=authorship,
            permissions=permissions,
            resource_id=self.res_id,
        )
        self.file_value = IIIFUri.new(value=iiif_uri, metadata=meta, comment=comment, resource_id=self.res_id)
        return self

    #######################
    # Migration Metadata
    #######################

    def add_migration_metadata(
        self, creation_date: str | None, iri: str | None = None, ark: str | None = None
    ) -> Resource:
        """
        Add metadata from a SALSAH migration.

        [See XML documentation for details](https://docs.dasch.swiss/latest/DSP-TOOLS/file-formats/xml-data-file/#describing-resources-with-the-resource-element)

        Args:
            creation_date: creation date of the resource in SALSAH
            iri: Original IRI in SALSAH
            ark: Original ARK in SALSAH

        Raises:
            XmllibInputError: if metadata already exists

        Returns:
            The original resource, with the added migration metadata

        Examples:
            ```python
            resource = resource.add_migration_metadata(
                iri="http://rdfh.ch/4123/DiAmYQzQSzC7cdTo6OJMYA",
                creation_date="1999-12-31T23:59:59.9999999+01:00"
            )
            ```

            ```python
            resource = resource.add_migration_metadata(
                ark="ark:/72163/4123-43xc6ivb931-a.2022829",
                creation_date="1999-12-31T23:59:59.9999999+01:00"
            )
            ```
        """
        if self.migration_metadata:
            msg_info = MessageInfo(
                "This resource already contains migration metadata, no new data can be added.", resource_id=self.res_id
            )
            raise_xmllib_input_error(msg_info)
        self.migration_metadata = MigrationMetadata(creation_date=creation_date, iri=iri, ark=ark, res_id=self.res_id)
        return self



# mypy: disable-error-code="no-untyped-def"

from pathlib import Path

import pytest
import regex
from lxml import etree

from dsp_tools.error.exceptions import InputError
from dsp_tools.utils.xml_parsing.parse_clean_validate_xml import _reformat_error_message_str
from dsp_tools.utils.xml_parsing.parse_clean_validate_xml import parse_and_clean_xml_file
from dsp_tools.utils.xml_parsing.parse_clean_validate_xml import parse_and_validate_xml_file


@pytest.fixture
def data_systematic_unclean() -> Path:
    return Path("testdata/xml-data/test-data-systematic.xml")


@pytest.fixture
def data_systematic_cleaned(data_systematic_unclean: Path) -> etree._Element:
    return parse_and_clean_xml_file(data_systematic_unclean)


def _clean_resulting_tree(tree: etree._Element) -> str:
    cleaned_str = regex.sub("\n", "", etree.tostring(tree, encoding=str))
    return regex.sub(" +", " ", cleaned_str)


def test_parse_and_clean_xml_file_same_regardless_of_input(data_systematic_unclean: Path) -> None:
    from_tree = parse_and_clean_xml_file(data_systematic_unclean)
    from_file = parse_and_clean_xml_file(Path("testdata/xml-data/test-data-systematic.xml"))
    cleaned_from_file = _clean_resulting_tree(from_file)
    cleaned_from_tree = _clean_resulting_tree(from_tree)
    assert cleaned_from_file == cleaned_from_tree, (
        "The output must be equal, regardless if the input is a path or parsed."
    )


def test_comment_removal(data_systematic_unclean: Path, data_systematic_cleaned: etree._Element) -> None:
    data = etree.parse(data_systematic_unclean)
    comments_before = [e for e in data.iter() if isinstance(e, etree._Comment)]
    assert len(comments_before) > 0
    comments_after = [e for e in data_systematic_cleaned.iter() if isinstance(e, etree._Comment)]
    assert len(comments_after) == 0, (
        "properties that are commented out would break the the constructor of the class XMLProperty, "
        "if they are not removed in the parsing process"
    )


def test_validate_xml_data_systematic() -> None:
    assert parse_and_validate_xml_file(input_file="testdata/xml-data/test-data-systematic.xml")


def test_validate_xml_data_minimal() -> None:
    assert parse_and_validate_xml_file(input_file="testdata/xml-data/test-data-minimal.xml")


def test_validate_xml_invalid_resource_tag_line_twelve() -> None:
    expected = regex.escape(
        "The XML file cannot be uploaded due to the following validation error(s):\n"
        "    Line 12: Element 'resource', attribute 'invalidtag': The attribute 'invalidtag' is not allowed."
    )
    with pytest.raises(InputError, match=expected):
        parse_and_validate_xml_file(input_file="testdata/invalid-testdata/xml-data/invalid-resource-tag.xml")


def test_validate_xml_data_duplicate_iri() -> None:
    expected_msg = regex.escape(
        "The XML file cannot be uploaded due to the following validation error(s):\n"
        "    Line 19: Element 'resource': Duplicate key-sequence ['http://rdfh.ch/4123/54SYvWF0QUW6a'] "
        "in unique identity-constraint 'IRI_attribute_of_resource_must_be_unique'."
    )
    with pytest.raises(InputError, match=expected_msg):
        parse_and_validate_xml_file(input_file="testdata/invalid-testdata/xml-data/duplicate-iri.xml")


def test_validate_xml_duplicate_ark() -> None:
    expected_msg = regex.escape(
        "The XML file cannot be uploaded due to the following validation error(s):\n"
        "    Line 19: Element 'resource': Duplicate key-sequence ['ark:/72163/4123-31ec6eab334-a.2022829'] "
        "in unique identity-constraint 'ARK_attribute_of_resource_must_be_unique'."
    )
    with pytest.raises(InputError, match=expected_msg):
        parse_and_validate_xml_file(input_file="testdata/invalid-testdata/xml-data/duplicate-ark.xml")


def test_validate_xml_empty_label() -> None:
    expected_msg = regex.escape(
        "The XML file cannot be uploaded due to the following validation error(s):\n"
        "    Line 11: Element 'resource', attribute 'label': [facet 'minLength'] "
        "The value '' has a length of '0'; this underruns the allowed minimum length of '1'."
    )
    with pytest.raises(InputError, match=expected_msg):
        parse_and_validate_xml_file(input_file="testdata/invalid-testdata/xml-data/empty-label.xml")


def test_validate_xml_invalid_characters_in_resptr() -> None:
    expected_msg = regex.escape(
        "The XML file cannot be uploaded due to the following validation error(s):\n"
        "    Line 13: Element 'resptr': [facet 'pattern'] The value 'not|allowed|characters' "
        "is not accepted by the pattern "
    )
    with pytest.raises(InputError, match=expected_msg):
        parse_and_validate_xml_file(input_file="testdata/invalid-testdata/xml-data/invalid-resptr-characters.xml")


def test_beautify_err_msg() -> None:
    _match = (
        r"Line \d+: The resource ID 'res_1' is not valid\. IDs must be unique across the entire file\. "
        r"The function make_xsd_compatible_id\(\) assists you in creating IDs\.\n +"
        r"Line \d+: The resource ID 'res_2' is not valid\. IDs must be unique across the entire file\. "
        r"The function make_xsd_compatible_id\(\) assists you in creating IDs\."
    )
    with pytest.raises(InputError, match=_match):
        parse_and_validate_xml_file("testdata/invalid-testdata/xml-data/duplicate-res-id.xml")


class TestReformatErrorMessage:
    def test_empty_label(self):
        in_msg = (
            "Element '{https://dasch.swiss/schema}resource', attribute 'label': [facet 'minLength'] "
            "The value '' has a length of '0'; this underruns the allowed minimum length of '1'."
        )
        result = _reformat_error_message_str(in_msg, line_number=1)
        assert result.line_number == 1
        assert result.element == "resource"
        assert result.attribute == "label"
        assert result.message == "The value '' has a length of '0'; this underruns the allowed minimum length of '1'."

    def test_pattern_resource_id(self):
        in_msg = (
            "Element '{https://dasch.swiss/schema}resptr': [facet 'pattern'] "
            "The value 'not|allowed|characters' is not accepted by the pattern "
            "'([a-zA-Zçéàèöäüòôûâêñ_][a-zA-Zçéàèöäüòôûâêñ_]*)'."
        )
        result = _reformat_error_message_str(in_msg, line_number=1)
        assert result.line_number == 1
        assert result.element == "resptr"
        assert result.attribute is None
        assert result.message == "The value 'not|allowed|characters' is not accepted by the pattern for this value."

    def test_invalid_tag(self):
        in_msg = (
            "Element '{https://dasch.swiss/schema}resource', attribute 'invalidattrib': "
            "The attribute 'invalidattrib' is not allowed."
        )
        result = _reformat_error_message_str(in_msg, line_number=1)
        assert result.line_number == 1
        assert result.element == "resource"
        assert result.attribute == "invalidattrib"
        assert result.message == "The attribute 'invalidattrib' is not allowed."

    def test_duplicate_iri(self):
        in_msg = (
            "Element '{https://dasch.swiss/schema}resource': "
            "Duplicate key-sequence ['http://rdfh.ch/4123/54SYvWF0QUW6a'] in unique identity-constraint "
            "'{https://dasch.swiss/schema}IRI_attribute_of_resource_must_be_unique'."
        )
        result = _reformat_error_message_str(in_msg, line_number=1)
        assert result.line_number == 1
        assert result.element == "resource"
        assert result.attribute is None
        assert (
            result.message == "Duplicate key-sequence ['http://rdfh.ch/4123/54SYvWF0QUW6a'] "
            "in unique identity-constraint 'IRI_attribute_of_resource_must_be_unique'."
        )

    def test_duplicate_ark(self):
        in_msg = (
            "Element '{https://dasch.swiss/schema}resource': Duplicate key-sequence "
            "['ark:/72163/4123-31ec6eab334-a.2022829'] in unique identity-constraint "
            "'{https://dasch.swiss/schema}ARK_attribute_of_resource_must_be_unique'."
        )
        result = _reformat_error_message_str(in_msg, line_number=1)
        assert result.line_number == 1
        assert result.element == "resource"
        assert result.attribute is None
        assert (
            result.message == "Duplicate key-sequence ['ark:/72163/4123-31ec6eab334-a.2022829'] "
            "in unique identity-constraint 'ARK_attribute_of_resource_must_be_unique'."
        )

    def test_resource_id_problem(self):
        in_msg = (
            "Element '{https://dasch.swiss/schema}audio-segment', attribute 'id': "
            "'res_1' is not a valid value of the atomic type 'xs:ID'."
        )
        result = _reformat_error_message_str(in_msg, line_number=1)
        assert result.line_number == 1
        assert result.element == "audio-segment"
        assert result.attribute == "id"
        assert (
            result.message == "The provided resource id 'res_1' is either not a valid xsd:ID or not unique in the file."
        )


if __name__ == "__main__":
    pytest.main([__file__])
